"""
Exceptions
^^^^^^^^^^

Exceptions used by the quality module.
"""

################################
# Exceptions
################################

class NotInitialized(Exception):
    """Initialize to discover all available expectation
    implementations. It has not happened.
    """
    pass

class InvalidExpectation(Exception):
    """
    Invalid specification of expectations
    """
    pass

class NoExpectations(Exception):
    """
    Empty list of expectations to evaluate
    """
    pass

class UnsupportedExpectation(Exception):
    """
    Expectation name or content not implemented/supported.
    """
    pass


class IncorrectImplementationExpectation(Exception):
    """
    Expectation implementation class didnt return right result
    """
    pass

class InvalidConfigurationExpectation(Exception):
    """
    Expectation implementation class didnt return right result
    """
    pass



import os
import sys
import json
import copy
import re
import traceback 
import logging
import markdown
from datetime import timezone
import pytz 
import humanize 
from ..lib.exceptions import * 
from .node import *

logger = logging.getLogger('app')

__all__ = ['SearchSkin', 'GenericSkin']

class SearchSkin(Skin): 
    """

    This provides a search interface to the dataframes. 

    .. deprecated:: 
           Use enrichsdk.render functions from the app 

    """
    def __init__(self, *args, **kwargs):
        super(SearchSkin, self).__init__(*args, **kwargs) 

        self.framename = None 
        """
        Dataframe to service the search interface 
        """

        self.autocomplete_column = None
        """
        Column to feed the autocomplete function of the search interface
        """

        self.dashboard_rules = [] 
        """
        Specification of the layout of the search result. 
        """
        self.run = None 
        self.detail = None 
        self.tags.extend(['skin', 'search']) 

    def load(self): 

        # If not loaded or loaded run not the last one: 
        lastrun = self.config.runmanager.get_last_run()
        if ((self.run is not None) and 
            (lastrun is not None) and 
            (lastrun.runid == self.run.runid)):
            return 

        # Note it 
        self.run = lastrun 

        # If still not loaded, give up 
        if self.run is None: 
            return

        # Load the details...
        self.run.load_files(include_output_frame=True) 
        self.detail = self.run.find_file_detail(framename=self.framename, 
                                                action='output')

        if self.detail is None: 
            raise Exception("Could not find detail") 

    def autocomplete(self, term): 

        if self.run is None: 
            try: 
                self.load() 
            except: 
                pass 

        if self.run is None or self.detail is None: 
            return [] 

        df = self.detail['df'] 
        names = df[self.autocomplete_column].values 
        names = [n for n in names if term.lower() in n.lower()]

        data = [{
            'label': n,
            'key': n
        } for n in names][:10]

        return data 

        
    def render_helper(self, rules, row):     
        data = []
        html = ""
        for r in rules: 
            enable = r.get('enable', True) 
            if not enable: 
                continue 
            c = r['context'](row) 
            rendered = self.template_render(r['template'], c)
            html += "\n" + rendered 

        data.append({
            'target': 'result', 
            'rendered': html
        })

        return data 

    def render(self, name):

        if self.run is None: 
            try: 
                self.load() 
            except: 
                pass 

        if self.run is None: 
            return {}

        if name is None or len(name) == 0: 
            return {} 

        if self.detail is None:
            # Try loading 
            try: 
                self.load() 
            except: 
                pass 

        if self.detail is None:             
            return {} 

        df = self.detail['df']
        row = df[df[self.autocomplete_column] == name]

        if len(row) > 0: 
            row = row.iloc[0].to_dict()
            data = self.render_helper(self.dashboard_rules, row)
        else: 
            data = []

        return data

class GenericSkin(Skin): 
    """
    This skin allows rendering of arbitrary content. 
    
    It takes all available dataframes in the last available run as
    input. The render function can display content from one or more
    dataframes.
    """
    def __init__(self, *args, **kwargs):
        super(GenericSkin, self).__init__(*args, **kwargs) 
        self.run = None     
        self.tags.extend(['skin', 'generic']) 

    def load(self): 
        """
        Load the last run of the pipeline that this skin is 
        included in.

        """
        # If not loaded, load now 
        lastrun = self.config.runmanager.get_last_run()
        if ((self.run is not None) and 
            (lastrun is not None) and 
            (lastrun.runid == self.run.runid )): 
            return 

        # Note it 
        self.run = lastrun 
        
        # If still not loaded, give up 
        if self.run is None: 
            logger.error("Could not find run in GenericSkin")
            return

        # Load the details...
        self.run.load_files(include_output_frame=True) 

    def discover_params_helper(self, data):
        """
        This is overridden by the renderer 
        """
        raise Exception("To be overridden by the renderer") 

    def discover_params(self): 
        """
        Parameters that can be passed 
        
          {
             name: [val1, val2, ..]
          }

        """
        
        if self.run is None: 
            try: 
                self.load() 
            except: 
                traceback.print_exc() 
                pass

        if self.run is None: 
            raise Exception("Run not available") 

        # Collect all the frames...
        data = self.run.get_file_hash() 

        return self.discover_params_helper(data) 

        
    def render_helper(self, rules, params): 
        """
        This function renders the last run's data according
        to rules. 
    
        Args: 
            rules (list): A list of widget configurations 

        Returns: 

            Rendered html

        """
        # Collect all the frames...
        data = self.run.get_file_hash() 

        response = []
        html = ""
        for r in rules: 
            enable = r.get('enable', True) 
            meta = r.get('meta', False) 
            if not enable: 
                continue 

            t = r['template']
            if not meta: 
                c = r['context'](data, params) 
                rendered = self.template_render(t, c)
                html += "\n" + rendered 
            else: 
                
                # Render the subset of rules...
                subresponse = self.render_helper(r['rules'], params) 
                subresponsehtml = "\n".join([s['rendered'] for s in subresponse])
                
                # Now incorporate it...
                c = r['context'](data, params, subresponsehtml)
                rendered = self.template_render(t, c)
                html += "\n" + rendered 

        response.append({
            'target': 'result', 
            'rendered': html
        })

        return response 

    def render(self, params):
        """
        Generic render function that first loads the run, 
        all the files that are part of the run, and call 
        the render helper. 
        """
        try: 
            self.load() 
        except: 
            traceback.print_exc() 
            pass 

        if self.run is None: 
            return {}

        data = self.render_helper(self.dashboard_rules, params)

        return data 

    def get_metadata_context(self, data, params): 
        """
        Include standard metadata 
        """

        pairs = [ 
            {
                'name': 'RunID', 
                'value': self.run.runid
             },
            {
                'name': 'Run Date',
                'value': humanize.naturaltime(self.run.start_time.replace(tzinfo=None))
            }
        ]

        return {
            'title': 'Run Metadata', 
            'width': 12,
            'pairs': pairs
        }



import os
import copy
import json
import logging
from datetime import datetime
from dateutil.relativedelta import relativedelta
from django.urls import reverse
import sqlite3

from enrichsdk.lib.customer import find_usecase

from .models import ComplexSearchRecord

logger = logging.getLogger('app')

dbpath = os.path.expandvars("path-to-db.sqlite")

def get_complex_task_url(spec, request):
    return reverse("APPNAME:complex_request")

def get_filters():
    """
    All the Filters that must could be applied
    """
    if not os.path.exists(dbpath):
        return []

    # First get all the rules...
    con = sqlite3.connect(dbpath)
    cur = con.cursor()

    # => Get all the rules...
    table = "some_table"
    rules = list(cur.execute(f'SELECT distinct rule FROM {table} order by rule'))
    rules = [r[0] for r in rules]
    rule_conditions = []
    for index, r in enumerate(rules):
        rule_conditions.append({
            "value": r,
            "condition": f"(rule = '{r}')"
        })

    result = [
        {
            "name": "Matching Rule",
            "conditions": rule_conditions
        }
    ]

    con.close()

    return result

def get_queries():

    base = """\
SELECT *
FROM {}
WHERE (%(conditions)s)
ORDER BY total_txn_value desc
limit {}"""

    # Note: We are implementing uniqueness of the kycid using the groupby
    return [
        {
            "name": "All Available",
            "sql": base.format('tablename', 50000),
            "clean": lambda df: df.groupby('primary_key_id').first().reset_index()
        },
        {
            "name": "Top 10",
            "sql": base.format("tablename", 100),
            "clean": lambda df: df.groupby('primary_key_id').first().reset_index()[:10]

        },
        {
            "name": "Top 50",
            "sql": base.format("tablename", 500),
            "clean": lambda df: df.groupby('primary_key_id').first().reset_index()[:50]
        }
    ]

search_spec = {
    'name': "Profiles",
    'description': "Profiles for Multiple Usecases",
    'usecase': find_usecase(__file__),
    'links': [
        {
            "name":"APPNAME",
            "url": "APPNAME:index"
        }
    ],
    "models": {
        "search_record": ComplexSearchRecord
    },
    "personas": [
        {
            "type": "sqlite",
            "dbpath": dbpath,
            "name": "Profiles",
            "description": "Profiles match one of many rules. Note that there could be multiple entries per",
            "tables": [
                {
                    "name": "Sample",
                    "table": "tablename",
                    "action": "sample",
                    "download": True,
                },
                {
                    "name": "Queries",
                    "table": "tablename",
                    "action": "query",
                    "download": True,
                    "queries": get_queries(),
                    "filters": get_filters(),
                    "workflow_columns":[
                        'rule',
                        'columns_to_post_to_workflow'
                    ]
                },
                {
                    "name": "Search",
                    "table": "tablename",
                    "action": "search",
                    "search_columns":[
                        'columns_to_search'
                    ],
                    "workflow_columns":[
                        'rule',
                        'columns_to_post_to_workflow'
                    ]
                }
            ],
            "workflows": [
                {
                    "name": "ComplexTask",
                    "url": get_complex_task_url
                },
            ]
        },
    ]
}

def get_spec():
    spec = copy.copy(search_spec)
    return spec



import os
from prompt_toolkit.validation import Validator, ValidationError

#################################################
# Validators
#################################################
class NameValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(message='Input should not be empty',
                                  cursor_position=len(document.text)) # Move cursor to end of input.

class RealtimeBatchValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(message='Computation should not be empty',
                                  cursor_position=len(document.text))
        if document.text not in ['batch', 'realtime']:
            raise ValidationError(message="Valid values are 'batch' and 'realtime'",
                                  cursor_position=len(document.text))
class EmailValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(message='Email should not be empty',
                                  cursor_position=len(document.text))
        if "@" not in document.text:
            raise ValidationError(message='Invalid email address',
                                  cursor_position=len(document.text))


class OrgDescValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(message='Customer organization description should not be empty',
                                  cursor_position=len(document.text)) # Move cursor to end of input.

class GenericDescValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(message='Customer organization description should not be empty',
                                  cursor_position=len(document.text)) # Move cursor to end of input.

class LogoValidator(Validator):
    def validate(self, document):
        url = document.text
        if len(url) == 0:
            raise ValidationError(message='Customer logo URL should not be empty',
                                  cursor_position=len(document.text)) # Move cursor to end of input.
        if not url.startswith('http'):
            raise ValidationError(message='Customer logo URL should be a valid URL (http://...',
                                  cursor_position=len(document.text)) # Move cursor to end of input.
        if not (url.lower().endswith('png') or
                url.lower().endswith('gif') or
                url.lower().endswith('jpg') or
                url.lower().endswith('ico')):
            raise ValidationError(message='Customer logo URL should be a valid URL (http://....png',

                                  cursor_position=len(document.text)) # Move cursor to end of input.


class URLValidator(Validator):
    def validate(self, document):
        url = document.text
        if len(url) == 0:
            raise ValidationError(message='URL should not be empty',
                                  cursor_position=len(document.text)) # Move cursor to end of input.
        if not url.startswith('http'):
            raise ValidationError(message='URL should be a valid URL (http://...',
                                  cursor_position=len(document.text)) # Move cursor to end of input.


class ValidDir(Validator):
    def validate(self, document):
        if ((len(document.text) == 0) or
            (not os.path.exists(document.text)) or
            (not os.path.isdir(document.text))):
            raise ValidationError(message='Directory specified doesnt exist or is invalid',
                                  cursor_position=len(document.text))

class ValidFile(Validator):
    def validate(self, document):
        if ((len(document.text) == 0) or
            (not os.path.exists(document.text)) or
            (not os.path.isfile(document.text))):
            raise ValidationError(message='File specified doesnt exist or is invalid',
                                  cursor_position=len(document.text))




#!/usr/bin/env python3

# Core
import os
import sys
import pip
import base64
import subprocess
import json
import imp
import yaml
import traceback
import pkg_resources
import logging, traceback
import pyfiglet

# thirdparty
import click
from pythonjsonlogger import jsonlogger

# self
import enrichsdk
from enrichsdk import package, datasets, api as sdkapi, lib, featurestore as fslib, utils

@click.group()
def process():
    """
    Bootstrap/test/install Enrich modules and access server

    \b
    Getting started:
       version:  Version of this sdk
       start:    First time instructions
       env:      Setup the environment context

    \b
    Development:
       bootstrap:  Bootstrap modules including transforms
       test:      Test transforms, manage datasets

    \b
    Server:
       api:       Access the server API

    \b
    Utils:
       sample:    Sample data for sharing

    \b
    Helpers:
       show-log:  Pretty print log output
    """
    pass

@click.command('version', hidden=True)
def _version():
    """
    Show version
    """
    print(enrichsdk.__version__)

@click.command('start', hidden=True)
@click.option('--commands/--no-commands',
              default=False,
              help="Show commands")
def _start(commands):
    """
    Instructions to get started
    """

    result = pyfiglet.figlet_format("Enrich SDK", font = "slant"  )
    print(result)

    home = os.environ['HOME']
    if "ENRICH_ROOT" in os.environ:
        enrich_root = os.environ['ENRICH_ROOT']
    else:
        enrich_root = "$ENRICH_ROOT"

    if 'VIRTUAL_ENV' in os.environ:
        venv_root = os.environ['VIRTUAL_ENV']
        venv_name = os.path.basename(venv_root)
        install_settings = [
            'echo "#source {}/bin/activate" > $ENRICH_ROOT/env.sh'.format(venv_root),
            'echo "workon {}" >> $ENRICH_ROOT/env.sh'.format(venv_name),
        ]
    elif 'CONDA_PREFIX' in os.environ:
        venv_root = os.environ['CONDA_PREFIX']
        venv_name = os.path.basename(venv_root)
        install_settings = [
            'echo "conda activate {}" >> $ENRICH_ROOT/env.sh'.format(venv_name),
        ]
    else:
        print("Please use virtualenvwrapper or conda to install sdk")
        return



    steps = [
        {
            "description": "Understand Enrich and SDK",
        },
        {
            "description": "set ENRICH_ROOT and populate",
            "cmd": [
                "export ENRICH_ROOT={}/work/enrich".format(home),
                "mkdir -p $ENRICH_ROOT",
                "enrichpkg env populate"
            ]
        },
        {
            "description": "Check and update siteconf and versionmap",
            "cmd": [
                "enrichpkg env check",
                "cat $ENRICH_ROOT/etc/siteconf.json",
            ]
        },
        {
            "description": "[OPTIONAL] For complex deployments, create an environment/context file",
            "cmd": "enrichpkg env sample-context > context.yaml",
        },

        {
            "description": "[OPTIONAl] Create a simple settings file",
            "cmd": [
                '# file to be sourced before you start working. use the appropriate',
                '# environment activation mechanism'
            ] + install_settings + [
                'echo "export ENRICH_ROOT={}" >> $ENRICH_ROOT/env.sh'.format(enrich_root)
            ]
        },
        {
            "description": "Change dir to $ENRICH_CUSTOMERS",
            "cmd": "cd $ENRICH_ROOT/customers",
        },
        {
            "cmd": [
                "#To handle python paths etc. avoid spaces and hyphen in the names",
                "git checkout git@github.com:alphainc/enrich-acme.git acme"
            ],
            "description": "GIT checkout a code repository"
        },
        {
            "cmd": "cd $ENRICH_ROOT/customers/acme",
            "description": "Change to checked out repository"
        },
        {
            "cmd": [
                "enrichpkg bootstrap -c repo -p ."
            ],
            "description": "Bootstrap the repo if not already done"
        },
        {
            "cmd": [
                "# Typically usecases have first letter in capital",
                "enrichpkg bootstrap -c usecase -p Marketing"
            ],
            "description": "Create usecase, say Marketing"
        },
        {
            "cmd": [
                "cd Marketing"
            ],
            "description": "Change to usecase"
        },
        {
            "description": "Bootstrap a transform",
            "cmd": [
                "# Will create a python script",
                "enrichpkg bootstrap -c transform-simple -p transforms/helloworld.py",
                "# Will create a python module",
                "enrichpkg bootstrap -c transform-simple -p transforms/helloworld",
                "# Will create a python package",
                "enrichpkg bootstrap -c transform-package -p transforms/helloworld",
                "# Will create a hello world script",
                "enrichpkg bootstrap -c transform-helloworld -p transforms/helloworld.py",
            ]
        },
        {
            "cmd": [
                "# Add any requirements and install them",
                "vi ../requirements.txt",
                "pip install -r ../requirements.txt",
            ],
            "description": "Install repo-specific requirements"
        },
        {
            "cmd": [
                "enrichpkg test transform transforms/helloworld.py",
                "# To capture all debug logs",
                "enrichpkg test transform --capture transforms/helloworld.py",
            ],
            "description": "Test the transform"
        },
        {
            "cmd": [
                "enrichpkg bootstrap -c pipeline-conf -p pipelines/conf/helloworld.py"
            ],
            "description": "Bootstrap a pipeline"
        },
        {
            "cmd": [
                "enrichpkg test conf --capture pipelines/conf/helloworld.py"
            ],
            "description": "Check pipeline"
        },
        {
            "cmd": [
                "# prefect is the default workflow engine",
                "enrichpkg bootstrap -c prefectjob -p workflows/prefect/daily.py"
            ],
            "description": "Bootstrap a Prefect workflow"
        },
        {

            "cmd": [
                "workflows/prefect/daily.py"
            ],
            "description": "Check workflow"
        },
    ]

    if not commands:
        print("Note: Use --commands to see how to implement these steps\n")

    for i, s in enumerate(steps):
        print("[{:2}] {}".format(i+1, s['description']))
        if not commands:
            continue

        print("")
        if 'cmd' in s:
            if isinstance(s['cmd'], str):
                print("    $ {}".format(s['cmd']))
            else:
                for c in s['cmd']:
                    print("    $ {}".format(c))
        print("")


@click.command('bootstrap', hidden=True)
@click.option("--component", '-c',
              required=True,
              help="What should be bootstrapped",
              type=click.Choice(['repo',
                                 'usecase',
                                 'transform-package',
                                 'transform-simple',
                                 'transform-helloworld',
                                 'transform-query',
                                 'asset',
                                 'dashboard',
                                 'task-lib',
                                 'task-conf',
                                 'pipeline-conf',
                                 'sparkjob',
                                 #'airflow',
                                 'prefectjob',
                                 'rscript',
                                 'pyscript']))
@click.option("--path", '-p',
              required=True,
              help="Directory where the files should be stored")
@click.option("--template", '-t',
              required=False,
              default=None,
              help="Template to override the defaults")
@click.option('--context',
              default=None,
              help="Environment file")
def bootstrap_(component, path, template, context):
    """
    Bootstrap a fresh module
    """

    """
    Modules currently supported include transform, model, and skin.

    For now only the transforms and models are testable locally
    without the need for a test server.

    In future we expect to more comprehensively support the
    development of all modules.

    A custom template can be provided as well.
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise
    context.set_env()

    if not package.checkenv(validate_only=True):
        print("Environment is incomplete")
        print("Try: enrichpkg env check")
        return

    package.bootstrap(component, path, template)

@click.command('show-log', hidden=True)
@click.argument('logfile')
def _show_log(logfile):
    """
    Pretty-print run log
    """

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.print_log(logfile)

@click.group('env', hidden=True)
@click.pass_context
def envsetup(ctx):
    """
    Setup the environment. Environment includes:\n
      \b
      (a) Workspace for enrich to work
      (b) Minimal configuration

    This can be specified using environment variables
    and/or 'context' file.

    The absolute minimal required environment variable
    is ENRICH_ROOT (path to enrich workspace e.g.,
    ~/enrich)
    """

    ctx.ensure_object(dict)

@envsetup.command('sample-context')
@click.option("--root", '-r',
              default=None,
              help="Root of enrich")
def _context_generate(root):
    """
    Generate sample context file
    """

    if root is not None:
        os.environ['ENRICH_ROOT'] = os.path.abspath(root)
    elif 'ENRICH_ROOT' not in os.environ:
        os.environ['ENRICH_ROOT'] = os.path.expandvars("$HOME/enrich")

    context = lib.Context()
    context = context.asdict()

    if 'siteconf' not in context:
        context['siteconf'] = os.path.join(context['ENRICH_ETC'],
                                           'siteconf.json')

    print("Please store this in context.yaml")
    print("Minimum required valid paths include: siteconf, versionmap, enrich_data")
    print("Example: enrichpkg api --context context.yaml ...")
    print("----")
    print(yaml.dump(context, indent=4))

@envsetup.command('sample-versionmap')
def _versionmap():
    """
    Generate sample versionmap
    """

    print("Please store this in versionmap.json. Default is $ENRICH_ROOT/etc/versionmap.json")
    print("Update context.yaml with this versionmap path")
    print("----")
    print(package.get_sample_versionmap())

@envsetup.command('sample-siteconf')
def _siteconf():
    """
    Generate sample siteconf
    """

    print("Please store this in siteconf.json. Default is $ENRICH_ROOT/etc/siteconf.json")
    print("Update context.yaml with this siteconf path")
    print("----")
    print(package.get_sample_siteconf())

@envsetup.command('populate')
@click.option("--context",
              default=None,
              help="Context file")
def _populate(context):
    """
    Populate the directories
    """
    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        print("Error! " + str(e))
        return

    context.set_env()

    package.populate(context=context.asdict())

@envsetup.command('check')
@click.option("--context",
              default=None,
              help="Context file")
def _checkenv(context):
    """
    Check environment
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        print("Error! " + str(e))
        return

    context.set_env()

    package.checkenv(context=context)


##########################################################
# Test command
##########################################################
@click.group('test', hidden=True)
@click.option('--context',
              default=None,
              help="Environment file")
@click.pass_context
def test(ctx, context):
    """
    Test transforms/pipelines etc.
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise

    context.set_env()
    ctx.ensure_object(dict)
    ctx.obj['context'] = context.asdict()


@test.command('transform')
@click.argument("pkgdir")
@click.option('--capture/--no-capture',
              default=True,
              help="Capture output")
@click.pass_context
def test_pkg(ctx, pkgdir, capture):
    """
    Unit testing of a package module (transform)
    """
    context = ctx.obj['context']

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.test_transform(pkgdir, capture, context)

@test.command('task-lib')
@click.argument("taskdir")
@click.option('--capture/--no-capture',
              default=True,
              help="Capture output")
@click.pass_context
def test_task(ctx, taskdir, capture):
    """
    Unit testing of a task library
    """

    context = ctx.obj['context']

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.test_task(taskdir, capture, context)

@test.command('conf')
@click.argument("spec")
@click.option('--capture/--no-capture',
              default=True,
              help="Capture output")
@click.pass_context
def test_conf(ctx, spec, capture):
    """
    Minimal testing of pipeline/task configuration
    """

    context = ctx.obj['context']

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.test_conf(spec, capture, context)

@click.group("data")
@click.argument("spec")
@click.option('--capture/--no-capture',
              default=True,
              help="Capture output")
@click.pass_context
def testdata(ctx, spec, capture):
    """
    Manage test data


    spec could be transform or a spec file
    """
    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    ctx.ensure_object(dict)
    ctx.obj['spec'] = spec
    ctx.obj['capture'] = capture

@testdata.command('list')
@click.pass_context
def _testdata_list(ctx):
    """
    List available test datasets
    """

    package.testdata_action("list",
                            ctx.obj['spec'],
                            ctx.obj['capture'],
                            ctx.obj['context'])

@testdata.command('show')
@click.argument('dataset')
@click.pass_context
def _testdata_show(ctx, dataset):
    """
    Show the details of a given dataset
    """

    package.testdata_action("show",
                            ctx.obj['spec'],
                            ctx.obj['capture'],
                            ctx.obj['context'],
                            name=dataset)

@testdata.command('run',
                  context_settings=dict(
                      ignore_unknown_options=True,
                      allow_extra_args=True,
                  ))
@click.argument('dataset')
@click.argument('command')
@click.pass_context
def _testdata_action(ctx, dataset, command):
    """
    Generate action commands for a dataset over a specified range

    Pass additional parameters as key-value pairs as
    shown in show.
    """

    args = dict()
    for item in ctx.args:
        args.update([item.split('=')])

    package.testdata_action("action",
                            ctx.obj['spec'],
                            ctx.obj['capture'],
                            ctx.obj['context'],
                            name=dataset,
                            command=command,
                            **args)


test.add_command(testdata)

#########################################
# API
#########################################
@click.group(hidden=True)
@click.option("--config",
              default=None,
              help="API Config File")
@click.pass_context
def api(ctx, config):
    """
    Access Enrich Server
    """

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    alt = os.path.expandvars("$ENRICH_ETC/api.json")
    if config is None:
        config = alt

    # Write a default
    default = {
        "schema": "v1:api:enrich",
        "backends": [
        ]
    }
    if not os.path.exists(config):
        with open(config, 'w') as fd:
            fd.write(json.dumps(default, indent=4))

    ctx.ensure_object(dict)
    ctx.obj['config'] = config

####################################
# Support for sampling data
####################################
@click.group('sample', hidden=True)
@click.pass_context
def sample(ctx):
    """
    Sample data from blockstores
    """
    pass

@sample.command("crawl")
@click.argument("root")
@click.option("--ext", default="json",
              help="Extension of files to look for")
@click.pass_context
def crawl(ctx, root, ext):
    """
    Crawl the root to find possible files

    Args:
       root: Path in blockstore with s3:// prefix
    """

    utils.sample_crawl(root, ext)

@sample.command("cat")
@click.argument("path")
@click.pass_context
def crawl(ctx, path):
    """
    Show the contents of path

    Args:
       path: Path to file in blockstore with s3:// prefix
    """

    utils.sample_cat(path)

####################################
# Configure access to server
####################################
@click.group()
@click.pass_context
def config(ctx):
    """
    Configure access to Enrich Server
    """
    pass

@config.command("init")
@click.pass_context
def _config_init(ctx):
    """
    Initialize the config file
    """
    default = {
        "schema": "v1:api:enrich",
        "backends": [],
        "services": [],
    }
    filename = ctx.obj['config']
    with open(filename, 'w') as fd:
        fd.write(json.dumps(default, indent=4))

    print("Initialized config")

@config.command("list")
@click.pass_context
def _config_list(ctx):
    """
    List available backends/services
    """
    try:
        filename = ctx.obj['config']
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj['config']))

    print("Backends")
    print("-----")
    backends = config['backends']
    for b in backends:
        key = base64.b64decode(b['key'].encode('utf-8')).decode('utf-8')
        try:
            print("{} [{} @ {}]".format(b['name'], key,b['server']))
        except:
            pass

    print("\n")
    print("Services")
    print("-----")
    services = config.get('services',[])
    for s in services:
        try:
            print("{} [{}]".format(s['name'], s['path']))
        except:
            pass

@config.command("backend-add")
@click.argument("name")
@click.argument("server")
@click.argument("key")
@click.pass_context
def _backend_add(ctx, name, server, key):
    """
    Add backend
    """
    try:
        filename = ctx.obj['config']
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj['config']))

    # Encode the key before storing
    key = base64.b64encode(key.encode('utf-8')).decode('utf-8')

    backends = config['backends']
    missing = True
    for b in backends:
        if ('server' in b) and (b['server'] == server):
            b['name'] = name
            b['server'] = server
            b['key'] = key
            missing = False
            break

    if missing:
        backends.append({
            'name': name,
            'server': server,
            'key': key
        })

    with open(filename, 'w') as fd:
        fd.write(json.dumps(config, indent=4))

    print("Added {}".format(name))

@config.command("backend-remove")
@click.argument("name")
@click.pass_context
def _backend_remove(ctx, name):
    """
    Add backend
    """
    try:
        filename = ctx.obj['config']
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj['config']))

    backends = config['backends']
    missing = True
    updated = [b for b in backends if ((b.get('server',None) != name) and
                                       (b.get('name',None) != name))]

    if len(backends) == len(updated):
        raise Exception("Backend not found: {}".format(name))

    config['backends'] = updated

    with open(filename, 'w') as fd:
        fd.write(json.dumps(config, indent=4))

    print("Removed {}".format(name))

@config.command("service-add")
@click.argument("name")
@click.argument("path")
@click.pass_context
def _service_add(ctx, name, path):
    """
    Add URL
    """
    try:
        filename = ctx.obj['config']
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj['config']))

    if 'services' not in config:
        config['services'] = []

    services = config['services']
    missing = True
    for s in services:
        if ('name' in s) and (s['name'] == name):
            s['name'] = name
            s['path'] = path
            missing = False
            break

    if missing:
        services.append({
            'name': name,
            'path': path
        })

    with open(filename, 'w') as fd:
        fd.write(json.dumps(config, indent=4))

    print("Added {}".format(name))

@config.command("service-remove")
@click.argument("name")
@click.pass_context
def _service_remove(ctx, name):
    """
    Remove service
    """
    try:
        filename = ctx.obj['config']
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj['config']))

    services = config['services']
    missing = True
    updated = [s for s in services if s.get('name',None) != name]

    if len(services) == len(updated):
        raise Exception("Service not found: {}".format(name))

    config['services'] = updated

    with open(filename, 'w') as fd:
        fd.write(json.dumps(config, indent=4))

    print("Removed {}".format(name))


@click.group()
@click.argument("backend")
@click.pass_context
def show(ctx, backend):
    """
    Access a backend
    """
    try:
        filename = ctx.obj['config']
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj['config']))

    backends = config['backends']

    relevant = None
    for b in backends:
        if ((b.get('server',None) == backend) or
            (b.get('name',None) == backend)):
            relevant = b
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))

    # Decode the key before accessing
    key = base64.b64decode(relevant['key'].encode('utf-8')).decode('utf-8')
    relevant['key'] = key

    backend = sdkapi.Backend(relevant)
    ctx.obj['backend'] = backend


@show.command("status")
@click.option("--offset", default=0)
@click.option("--page", default=20)
@click.pass_context
def _status(ctx,offset,page):
    """
    Execution status
    """

    backend = ctx.obj['backend']
    result = backend.status(offset,page)

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return


    columns = ['usecase', 'nature', 'name', 'status', 'runid', 'start_time', 'duration']
    sdkapi.draw(backend, 'status',
                result, columns)

@show.command("usecases")
@click.pass_context
def _usecases(ctx):
    """
    Show available usecases
    """

    backend = ctx.obj['backend']
    result = backend.usecases()

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return


    columns = ['customer', 'name','description']
    sdkapi.draw(backend, 'usecases',
                result, columns)

@show.command("health")
@click.pass_context
def _health(ctx):
    """
    Show system health
    """

    backend = ctx.obj['backend']
    result = backend.health()

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return


    sdkapi.draw_health(backend, 'health', result)

@show.command("pipelines")
@click.argument('usecase')
@click.pass_context
def _pipelines(ctx, usecase):
    """
    Show available pipelines
    """

    backend = ctx.obj['backend']
    result = backend.pipelines(usecase)

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return

    data = result['data']
    for d in data:
        d['description'] = d['description'][:20]
        runs = sorted(d['runs'],reverse=True)
        if len(runs) > 0:
            d['lastrun'] = runs[0]
            d['runs'] = len(runs)
        else:
            d['lastrun'] = ""
            d['runs'] = 0

    columns = ['usecase', 'name','description', 'path', 'runs', 'lastrun']
    sdkapi.draw(backend, 'pipelines',  result, columns,
                {
                    'Usecase': usecase,
                })

@show.command("run-list")
@click.argument('usecase')
@click.argument('pipeline')
@click.pass_context
def _runs(ctx, usecase, pipeline):
    """
    Show available runs for pipeline
    """

    backend = ctx.obj['backend']
    result = backend.run_list(usecase, pipeline)

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return

    result['data'] = sorted(result['data'],
                            key=lambda r: r['start_time'],
                            reverse=True)

    columns = ['runid', 'status','start_time', 'end_time']
    sdkapi.draw(backend, 'run-list',  result, columns,
                {
                    'Usecase': usecase,
                    'Pipeline': pipeline
                })

@show.command("run-detail")
@click.argument('usecase')
@click.argument('pipeline')
@click.argument('runid')
@click.pass_context
def _run_detail(ctx, usecase, pipeline, runid):
    """
    Show detail of a given run
    """

    backend = ctx.obj['backend']
    result = backend.run_detail(usecase, pipeline, runid)

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return


    sdkapi.draw_run_detail(backend, 'run-detail', result,
                           {
                               'Usecase': usecase,
                               'Pipeline': pipeline,
                               'RunID': runid
                           })

@show.command("tasks")
@click.argument('usecase')
@click.pass_context
def _tasks(ctx, usecase):
    """
    Show available tasks
    """

    backend = ctx.obj['backend']
    result = backend.tasks(usecase)

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return

    data = result['data']
    for d in data:
        d['description'] = d['description'][:20]
        runs = sorted(d['runs'],reverse=True)
        if len(runs) > 0:
            d['lastrun'] = runs[0]
            d['runs'] = len(runs)
        else:
            d['lastrun'] = ""
            d['runs'] = 0

    columns = ['usecase', 'name','description', 'path', 'runs', 'lastrun']
    sdkapi.draw(backend, 'tasks',  result, columns,
                {
                    'Usecase': usecase,
                })

@show.command("task-run-list")
@click.argument('usecase')
@click.argument('task')
@click.pass_context
def _task_runs(ctx, usecase, task):
    """
    Show available runs for task
    """

    backend = ctx.obj['backend']
    result = backend.task_run_list(usecase, task)

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return

    result['data'] = sorted(result['data'],
                            key=lambda r: r['start_time'],
                            reverse=True)

    columns = ['runid', 'status','start_time', 'end_time']
    sdkapi.draw(backend, 'task-run-list',  result, columns,
                {
                    'Usecase': usecase,
                    'Task': task
                })

@show.command("task-run-detail")
@click.argument('usecase')
@click.argument('task')
@click.argument('runid')
@click.pass_context
def _task_run_detail(ctx, usecase, task, runid):
    """
    Show detail of a given run
    """

    backend = ctx.obj['backend']
    result = backend.task_run_detail(usecase, task, runid)

    status = result['status']
    if status in ['failure', 'error']:
        print("Failure while accessing backend")
        if 'error' in result:
            print("Reason:", result['error'])
        return


    sdkapi.draw_run_detail(backend, 'run-detail', result,
                           {
                               'Usecase': usecase,
                               'Task': task,
                               'RunID': runid
                           })


@click.group()
@click.argument("backend")
@click.argument("service")
@click.pass_context
def featurestore(ctx, backend, service):
    """
    Access featurestore at backend
    """
    try:
        filename = ctx.obj['config']
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj['config']))

    backends = config['backends']

    relevant = None
    for b in backends:
        if ((b.get('server',None) == backend) or
            (b.get('name',None) == backend)):
            relevant = b
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))

    # Decode the key before accessing
    key = base64.b64decode(relevant['key'].encode('utf-8')).decode('utf-8')
    relevant['key'] = key

    backend = sdkapi.Backend(relevant)
    ctx.obj['backend'] = backend

    # Lookup the service as well..
    services = config['services']
    relevant = None
    for s in services:
        if s.get('name',None) == service:
            relevant = s
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))
    ctx.obj['service'] = s

@featurestore.command("post")
@click.argument('filename')
@click.option("--debug/--no-debug",
              default=False)
@click.pass_context
def _post(ctx, filename, debug,):
    """
    Post json to server
    """

    backend = ctx.obj['backend']
    service = ctx.obj['service']

    response = fslib.post(backend=backend,
                          service=service,
                          filename=filename,
                          debug=debug)
    print(response)

@featurestore.command("generate")
@click.option("--debug/--no-debug",
              default=False)
@click.pass_context
def _generate(ctx, debug):
    """
    Generate sample files
    """

    backend = ctx.obj['backend']
    service = ctx.obj['service']

    response = fslib.generate(backend=backend,
                              service=service,
                              debug=debug)
    print(json.dumps(response, indent=4))

@featurestore.command("download")
@click.option('--featuregroup_id', default=None)
@click.option('--run_id', default=None)
@click.option("--debug/--no-debug",  default=False)
@click.pass_context
def _spec_download(ctx, featuregroup_id, run_id, debug,):
    """
    Download a featuregroup specification
    """

    if featuregroup_id is None and run_id is None:
        raise Exception("One of the featuregroup_id or run_id should be specified")

    backend = ctx.obj['backend']
    service = ctx.obj['service']

    response = fslib.download(backend=backend,
                              service=service,
                              featuregroup_id=featuregroup_id,
                              run_id=run_id,
                              debug=debug)
    print(json.dumps(response, indent=4))


@featurestore.command("search", context_settings=dict(
    ignore_unknown_options=True,
))
@click.option("--debug/--no-debug",  default=False)
@click.argument('params', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def _spec_search(ctx, debug, params):
    """
    Search for a feature spec or run
    """

    backend = ctx.obj['backend']
    service = ctx.obj['service']

    def error():
        print("Args should have attr=value format")
        print("Example: ....local spec search name=customer_persona")
        print("You can give django filter query name__icontains=persona")
        return

    for p in params:
        if p.count("=") != 1:
            return error()

    params = { p.split("=")[0]: p.split("=")[1] for p in params}
    if len(params) == 0:
        return error()

    response = fslib.search(backend=backend,
                            service=service,
                            params=params,
                            debug=debug)

    print(json.dumps(response, indent=4))

api.add_command(config)
api.add_command(show)
api.add_command(featurestore)

###################################
# Add to process
###################################
process.add_command(_version)
process.add_command(_start)
process.add_command(bootstrap_)
process.add_command(envsetup)
process.add_command(api)
process.add_command(test)
process.add_command(sample)
process.add_command(_show_log)

def main():
    process()




import os
import sys
import json
import logging
import traceback

logger = logging.getLogger('app')

def should_skip(name):

    return (name.startswith(".") or
            name == "__pycache__")

def load_customer_assets(context={}):
    """
    Go through the customer directories and add the assets from all
    installed customer modules.

    The ENRICH_CUSTOMERS looks like system path, and it is processed
    as such. It gives priority to the modules earlier in the path
    than later. Typically you create a run directory and prepend
    that run directory in front of the standard directory.
    """

    assets = {}

    # Check if the environment is correctly configured
    enrich_customers = context.get('ENRICH_CUSTOMERS',
                                   os.environ.get('ENRICH_CUSTOMERS',
                                                  None))

    if enrich_customers is None:
        enrich_root = context.get('ENRICH_ROOT',
                                  os.environ.get('ENRICH_ROOT',
                                                 None))
        if enrich_root is not None:
            enrich_customers = os.path.join(enrich_root, "customers")


    if enrich_customers is None:
        logger.error("Customer root is missing : {}".format(enrich_customers))
        return assets

    #=> Collect all possible paths..
    customer_roots = {}
    for customer_root in enrich_customers.split(";"):
        if not os.path.exists(customer_root):
            continue
        for customer in os.listdir(customer_root):
            if customer not in customer_roots:
                customer_roots[customer] = os.path.join(customer_root, customer)

    # Now go through each of them...
    for customer, customer_dir in customer_roots.items():

        # customer : scribble
        # customer_dir : $ENRICH_ROOT/customers/scribble

        # parent : $ENRICH_ROOT/customers
        customer_root = os.path.dirname(customer_dir)

        if not os.path.isdir(customer_dir) or should_skip(customer):
            continue

        enrichfile = os.path.join(customer_dir, 'enrich.json')
        if not os.path.exists(enrichfile):
            logger.debug("Missing enrichfile. Skipping customer: {}".format(customer))
            continue

        try:
            enrich = json.load(open(enrichfile))
        except:
            logger.debug("Invalid enrichfile. Skipping customer: {}".format(customer))
            continue

        if not enrich.get('enable', True):
            logger.debug("Disabled customer. Skipping customer: {}".format(customer))
            continue

        for usecase in os.listdir(customer_dir):
            """
            Each usecase within the customer organization
            """
            usecasedir = os.path.join(customer_dir, usecase)

            if not os.path.isdir(usecasedir) or should_skip(usecase):
                continue

            enrichfile = os.path.join(usecasedir, 'enrich.json')

            if not os.path.exists(enrichfile):
                logger.debug("Missing enrichfile. Skipping usecase: {}:{}".format(customer,usecase))
                continue

            try:
                enrich = json.load(open(enrichfile))
            except:
                logger.debug("Invalid enrichfile. Skipping usecase: {}:{}".format(customer,usecase))

            if not enrich.get('enable', True):
                logger.debug("Disabled org. Skipping usecase: {}:{}".format(customer, usecase))
                continue


            # => Check assets
            # Contrib
            #        /pkg
            #            /assets
            #        /assets

            for assetsdir in [os.path.join(usecasedir, 'pkg', 'assets'),
                              os.path.join(usecasedir, 'assets')]:
                if not os.path.exists(assetsdir):
                    continue

                for a in os.listdir(assetsdir):
                    if should_skip(a):
                        continue

                    for pkgdir in [ os.path.join(assetsdir, a, a),
                                    os.path.join(assetsdir, a, 'src', a)]:

                        if not os.path.exists(pkgdir):
                            continue

                        libdir = os.path.abspath(os.path.dirname(pkgdir))
                        if libdir not in sys.path:
                            sys.path.append(libdir)

                        assets[a] = {
                            'fullpath': os.path.join(libdir, a),
                            'relpath': os.path.relpath(os.path.join(libdir, a),
                                                       customer_root),
                            'organization': enrich['org']['name'],
                            'usecase': usecase,
                            'customer': customer
                        }

    return assets

def find_usecase(path):
    """
    Given any filepath, find the usecase corresponding to it.
    """
    try:

        # Look through the loaded usecases from settings.
        from django.conf import settings

        path = os.path.realpath(path)
        for usecase in settings.USECASES:
            usecase_root = os.path.realpath(usecase['root'])
            if  os.path.commonprefix([path, usecase_root]) == usecase_root:
                return usecase
    except:
        logger.exception("Could not find the usecase for the path")

    return None





"""
Code that goes along with the Airflow tutorial located at:
https://github.com/airbnb/airflow/blob/master/airflow/example_dags/tutorial.py
"""

import os
import sys 
from datetime import timedelta, datetime 
from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from airflow.operators.python_operator import PythonOperator
from datetime import datetime, timedelta

default_args = {
    'owner': 'airflow',
    'depends_on_past': False,
    'start_date': datetime(2018, 4, 4, 0, 0, 0),
    'email': ['support@scribbledata.io'],
    'email_on_failure': True,
    'email_on_retry': True,
    'retries': 1,
    'retry_delay': timedelta(minutes=5),
    'catchup': False, 
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}

dag = DAG('contrib-pipeline-v1', 
          schedule_interval=timedelta(hours=2),
          catchup=False,
          depends_on_past=False,
          default_args=default_args)

cli = os.path.join(os.environ['ENRICH_ROOT'],
                   'manage', 'bin', 'run_enrichcli.sh')

conf = os.path.join(os.environ['ENRICH_CUSTOMERS'], 
                    'scribble', 'Contrib', 'conf', 
                    'test.json') 

t1 = BashOperator(
    task_id='run_test_pipeline',
    bash_command='{} run-pipeline -c {}'.format(cli, conf),
    dag=dag)


t2 = EmailOperator(
    task_id="notification",
    to='support@scribbledata.io',
    subject='Completed task',
    html_content='Completed ',
    dag=dag
)

t2.set_upstream(t1)



__all__ = [
    'FileOperationsBase',
    'FileBasedQueryExecutorBase',
    'NotebookExecutorBase',
    'InMemoryQueryExecutorBase',
    'FeatureComputeBase',
    'MetricsBase'
]

from .fileops import *
from .filebased_query_executor import *
from .inmemory_query_executor import *
from .notebook_executor import * 
from .feature_compute import *
from .metrics import *



#!/usr/bin/env python

import os
import sys
import click
import glob
import json
import re
import getpass 
import traceback
import importlib
import string 
import types
import platform 
from datetime import datetime 
from collections import OrderedDict 
from time import strftime,localtime
import logging
from functools import update_wrapper

from enrichsdk.commands import Config, RunManager 
from enrichsdk.commands.helper import extend_run_commandset
from enrichsdk.commands import log

@click.group()
@click.pass_context
@click.option('--debug/--no-debug',
              default=False)
@click.option('--readonly/--no-readonly',
              default=False)
@click.option('--show-metadata/--no-show-metadata',
              default=False)
@click.option('-s', '--settings',
              help="configuration file", 
              default=None) 
def main(ctx, debug, readonly, show_metadata, settings): 
    """
    Managed command execution. 

    This is to ensure repeated adhoc commands have a place to 
    go to, and to log all such commands. Further the outputs 
    of the commands can be downloaded from the GUI. 
    """

    if settings is None:
        tries = [
            os.path.expanduser('~/.enrichcmd-settings.json'),
            '.enrichcmd-settings.json',
        ]

        if 'ENRICH_ETC' in os.environ:
            tries.append(os.path.join(os.environ['ENRICH_ETC'],
                                      'enrichcmd-settings.json'))
        
        for t in tries:
            if os.path.exists(t):
                settings = t
                break 

        # If nothing matches, the use the default
        if settings is None:
            settings = tries[0]
            
    # Setup logging...
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level)

    # => Load the global configuration..
    config = Config({
        'debug': debug,
        'readonly': readonly,
        'show_metadata': show_metadata, 
        'settings': settings
    })
    ctx.obj = config

    if ctx.invoked_subcommand not in ['configure']: 
        global _run 
        config.configure(settings)    
        extend_run_commandset(_run, config) 

    # Disable whatever initial logging is there...
    log.logging_clear()
    
        
        
@main.command('configure') 
@click.argument('action', type=click.Choice(['init', 'show']))
@click.pass_context
def _configure(ctx, action):
    """
    Initialize the enrichcmd-settings.json
    
    The settings is required to tell enrichcmd where to 
    pickup the commands etc. And more 
    """

    config = ctx.obj
    
    defaults = OrderedDict([
        ("enrich_root", "~/enrich"),
        ("runid", "run-%Y%m%d-%H%M%S-%(username)s"),
        ("rundir", "%(enrich_data)s/_commands/%(runid)s"),
        ("outputdir", "%(rundir)s/output"),
        ("log", "%(rundir)s/log.json"),
        ("metadata", "%(rundir)s/metadata.json"),
        ("libraries", [
            "%(enrich_root)s/customers/alpha/commands"
        ]),
        ("args", OrderedDict([
            ("command1", {
                "path": "%(enrich_data)s/acme/Project" 
            })
        ]))
    ])

    settings = config.settings 
    if action == 'init' and os.path.exists(settings):
        print("Error! Settings file {} exists. Not overwriting. Please remove".format(settings))
        return

    if action == 'show' and not os.path.exists(settings):
        print("Error! Settings file {} does not exists. Please create one using 'configure init'".format(settings))
        return    

    if action == 'init': 
        with open(settings, 'w') as fd:
            json.dump(defaults, fd, indent=4)
            print("Created a default settings ({}). Please edit.".format(settings))
    else:
        print(settings)
        print("=====================")
        content = open(settings).read()
        print(content) 


        
@main.command('history')
@click.option('-n', 
              default=10,
              help="How many entries to show")
@click.pass_obj
def _history(config,n):
    """
    Show history of past commands 
    """
    mgr = RunManager({
        'command_root': config.command_root
    })

    mgr.show_history(n) 

@main.group('run') 
def _run():
    """
    Auditable command execution framework 
    
    A number of commands are supported to process partial data.
    """
    pass 

                     
if __name__ == "__main__":
    main() 



import os
import copy
import json
import numpy as np
from hashlib import sha256
from datetime import datetime
import logging

from enrichsdk import Source

logger = logging.getLogger("app")

class TableSource(Source):
    """
    Load csv/other files into pandas dataframes.

    Parameters specific to this module include:

        * source: A dictionary of dataframe names and how to
          load them. It has a number of attributes:

            * type: Output type. Only 'table' value is
              supported for this option.
            * filename: Output filename. You can use default
              parameters such  runid
            * params: Params are arguments to [pandas read_csv](https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html)

    Example::

        ....
        "transforms": {
            "enabled": [
                {
                    "transform": "TableSink",
                    "args": {
                        "source": {
                            "article": {
                                "type": "file",
                                "filename": "%(data)s/ArticleData.csv",
                                "params": {
                                    "delimiter": "|",
                                    "dtype": {
                                        "sku": "category",
                                        "mc_code": "int64",
                                        "sub_class": "category",
                                        "priority": "float64"
                                        ...
                                    }
                                }
                            }
                        }
                      ...
                    }
                }
            ...
           ]
         }

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "TableSource"

    def preload_clean_args(self, args):
        """
        Clean when the spec is loaded...
        """

        # Backward compatability
        if 'source' not in args:
            args = {
                'source': args
            }

        args = super().preload_clean_args(args)

        # Sanity check...
        assert isinstance(args, dict)
        assert 'source' in args
        assert isinstance(args['source'], dict)

        for name, detail  in args['source'].items():

            # Insert the frame into the args for backward
            # compatability.
            if (("type" in detail) and
                ("frametype" not in detail) and
                (detail['type'] == 'table')):
                detail['frametype'] = 'pandas'

            if (("frametype" not in detail) or
                (detail['frametype'] != 'pandas')):
                logger.error("Invalid configuration. Only pandas table source supported by this source transform",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            if (('filename' not in detail) or
                (not isinstance(detail['filename'], str)) or
                ('params' not in detail) or
                (not isinstance(detail['params'], dict))):
                logger.error("Invalid args. Filename (string) and params (dict) are required",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            mapping = {
                'str': str,
                'float': np.float64,
                'float64': np.float64,
                'np.float64': np.float64,
                'np.int64': np.int64,
                'int': np.int64,
                'int64': np.int64,
                'datetime': datetime,
                'category': 'category'
            }

            #=> Materialize the path...
            detail['filename'] = self.config.get_file(detail['filename'])
            detail['root'] = self.config.enrich_data_dir
            params = detail['params']
            if 'dtype' in params:
                for attr in params['dtype']:
                    if params['dtype'][attr] in mapping:
                        params['dtype'][attr] = mapping[params['dtype'][attr]]
                    else:
                        params['dtype'][attr] = eval(params['dtype'][attr])

        return args

    def validate_args(self, what, state):
        """
        """
        assert isinstance(self.args, dict)
        assert 'source' in self.args
        for name, detail  in self.args['source'].items():
            assert (('frametype' in detail) and (detail['frametype'] == 'pandas'))
            assert 'filename' in detail
            assert 'params' in detail

    def sample_inputs(self):

        frame = self.config.get_dataframe('pandas')

        # Check if nothing has been specified
        if 'source' not in self.args:
            return []

        args = self.args['source']

        # Read 50 rows only...
        args = copy.deepcopy(args)
        for name in args:
            if 'params' not in args[name]:
                args[name]['params'] = {}
            args[name]['params']['nrows'] = 50

        # => Load the tables
        try:
            dfstates = frame.read(args, {})
        except:
            logger.exception("Error while loading the inputs")
            return []

        # => Generate the inputs for the caller...
        inputs = []
        for name in args:
            detail = dfstates[name]
            table = [list(frame.columns(detail['df']))] + \
                    frame.tolist(frame.head(detail['df'], 4))
            inputs.append({
                'name': name,
                'rows': frame.shape(detail['df'])[0],
                'columns': frame.shape(detail['df'])[1],
                'params': detail['params'],
                'table': table
            })

        return inputs

    def clean(self, state):
        """
        This is meant for subclass to do some additional processing.
        """
        pass

    def process(self, state):
        """
        Load file...
        """
        # Load all the dataframes. This will use the full enrich
        # deployment's beefed up read function.
        framecls = self.config.get_dataframe('pandas')
        source  = self.args['source']

        dfstates = framecls.read(source, {})
        for dfname, dfstate in dfstates.items():

            # => Insert column description
            columns = dfstate['params']['columns']
            for c in columns:
                columns[c]['description'] = self.get_column_description(dfname, c)

            params = dfstate['params']
            if 'filename' in params:
                filename = params['filename']
            elif 'filename' in source.get(dfname,{}):
                filename = self.args[dfname]['filename']
            else:
                filename = "Unknown"

            detail = {
                'df': dfstate['df'],
                'transform': self.name,
                'frametype': 'pandas',
                'params': [
                    params,
                    {
                        "type": "lineage",
                        "transform": self.name,
                        "dependencies": [
                            {
                                "type": "file",
                                "nature": "input",
                                "objects": [filename]
                            }
                        ]
                    }
                ],
                'history': [
                    {
                        'transform': self.name,
                        'log': "Read data using {}".format(framecls.__class__.__name__)
                    }
                ]
            }
            try:
                state.update_frame(dfname, detail, create=True)
            except:
                logger.exception("Unable to store state",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Error while loading")

        # Clean the uploaded data...
        self.clean(state)

    def validate_results(self, what, state):
        pass

provider = TableSource



"""
Standard transforms that can be directly included in 
any pipeline.
"""

__all__ = ['FileOperations',
           'JSONSink', 'JSONSource',
           'PQExport', 'SQLExport',
           'TableSink', 'TableSource']

from .fileops import *
from .jsonsink import *
from .jsonsource import *
from .pqexport import *
from .sqlexport import *
from .tablesink import *
from .tablesource import *



"""Modules to support feature engineering of objects (dictionaries). 

There are two base classes at the individual feature level and at a
feature set level. There is a compute function that iterates through
these for all the objects 
"""

import os
import json
import logging
from collections import defaultdict
from datetime import datetime, date, timedelta
from dateutil import parser as dateparser
import pandas as pd

logger = logging.getLogger("app")

__all__ = [
    'FeaturesetExtractorBase',
    'FeatureExtractorBase',
    'compute_features',
    'note'
]
def note(df, title):
    """
    Quick summary of a dataframe including shape, column, sample etc.

    Args:
       df (dataframe): Input dataframe
       title (str): Title

   Returns:
       str: A formatted text to be used for logging

    """
    msg = title + "\n"
    msg += "--------" + "\n"
    msg += "Timestamp: " + str(datetime.now()) + "\n"
    msg += "\nShape: "+ str(df.shape) + "\n"
    msg += "\nColumns: " + ", ".join(df.columns) + "\n"
    if len(df) > 0:
        msg += "\nSample:" + "\n"
        msg += df.sample(min(2, len(df))).T.to_string() + "\n" + "\n"
    msg += "\nDtypes" + "\n"
    msg += df.dtypes.to_string() + "\n"
    msg += "------" + "\n"
    return msg


class FeatureExtractorBase:
    """
    Extract a single feature

    Example::

        class SimpleFeatureExtractor(FeatureExtractorBase):

            def __init__(self, *args, **kwargs):
                self.name = "simple"

            def extract(self, name, data, key=None):
                if key is None:
                    key = name
                value = jmespath.search(key, data)
                return [{
                    'key': name,
                    'value': value
                }]
            """
    def __init__(self, *args, **kwargs):
        self.name = "base"

    def extract(self, name, data, key=None):
        """
        Given data and a name, generate some attributes. The
        return value should be a list of dictionaries

        Args:
           name (str): name of the feature
           data (dict): A dictionary
           key (str): Dictionary key potentially if name is not key

        Returns:
           list: List of dictionaries. Each dict has a "key" and "value
        """
        raise Exception("Not implemented")


class FeaturesetExtractorBase:
    """
    Compute a featureset - collection of features. To be used in conjunction
    with the FeatureCompute(outer) and FeatureExtractor (inner).
    We define and use multiple extractors::

        class CustomerFeaturesetExtractor(FeaturesetExtractorBase):
            '''
            Customer  timeseries featureset extractor
            '''
            def get_extractors(self):
                return {
                    "simple": SimpleFeatureExtractor(),
                }

            def get_specs(self):
                specs = [
                    {
                        "keys": ['days'],
                        "extractor": "simple"
                    },
                ]

                return specs

            def one_record(self, data):

                allfeatures = super().one_record(data)

                return allfeatures

            def clean(self, df):
                 df = df.fillna("")
                 return df

    """

    def get_extractors(self):
        """
        Returns a list of extractors. This is over-ridden in
        the subclass. Sample::

            return {
                 "simple": SimpleFeatureExtractor(),
            }

        Returns:
           dict: Dictionary of name -> extractor class instance

        """
        return {}

    def get_specs(self):
        """
        Returns a list of specifications. Each specification applies to
        one or more features. We specify a combination of keys
        in the input dictionary and a corresponding extractor. The keys
        could be a list or a dictionary.

        For example::

            [
                {
                    "keys": ['age', 'sex'],
                    "extractor": "simple",
                },
                {
                    "keys": {
                         'gender':  'sex',
                         'old': 'age'
                     },
                    "extractor": "simple",
                }
            ]
            """
        return []

    def one_record(self, data):
        """
        Process one record at a time. Pass it through the
        extractors, collect the outputs and return

        Args:
            data (dict): One record to process

        Returns:
            list: A list of dictionaries with features from this record

        Rough logic::

             get specs
             for each spec:
                     find extractor
                     find name and corresponding keys
                     newfeatures = call extractor(name, keys) for one row in data
                     collect new features

             collapse
             return one or more 'feature row(s)'
        """
        allfeatures = []

        extractors = self.get_extractors()
        specs = self.get_specs()
        for spec in specs:

            extractor = spec.get('extractor', 'default')
            extractor = extractors[extractor]

            if isinstance(spec['keys'], list):
                for key in spec['keys']:
                    if key not in data:
                        continue
                    features = extractor.extract(key, data)
                    allfeatures.extend(features)
            elif isinstance(spec['keys'], dict):
                for name, key in spec['keys'].items():
                    features = extractor.extract(name, data, key)
                    allfeatures.extend(features)

        return allfeatures

    def collate(self, features):
        """
        Combine a outputs of the extractors (each of which is a dictionary)
        into an object. It could be anything that the cleaner can handle.

        Args:
           features (list): List of features extracted by one_record

        Returns:
           object: Could be a combined dictionary/dataframe/other

        """
        return pd.DataFrame(features)

    def clean(self, df):
        """
        Clean the collated dataframe/list/other.

        Args:
           df (object): output of collate

        Returns:
           object: A cleaned collated object
        """
        return df

    def finalize(self, df, computed):
        """
        Take cleaned data and generate a final object such as a dataframe

        Args:
           df (object): output of collate
           computed (dict): featureset extractor name -> collated/cleaned object

        Returns:
           object: final data object

        """
        return df

    def document(self, name, df):
        """
        Document the dataframe generated. The default is
        to capture schema, size etc. Over-ride to extend this
        documentation.

        Args:
           df (object): output of collate
           name (str): name of the featureset extractor specification

        """
        if not isinstance(df, pd.DataFrame):
            logger.error("Unable to document. Unsupported data format. Override method in subclass")
        else:
            return note(df, getattr(self, 'name', self.__class__.__name__))


def compute_features(objects, extractors, read_object=None):
    """
    Compute the features 

    Args:
          objects (list): List of objects to process. Could be names
          extractors (dict): Name to extractor mapping
          read_object (method): Turn each object into a dict

    Returns:
        dict: name to dataframe mapping
    
    """

    featuresets = defaultdict(list)
    counts = defaultdict(int)
    invalid_objects = []
    
    # First turn it into a list...
    if isinstance(objects, dict):
        objects = [objects]
        
    for obj in objects:
        try:
            counts['obj_total'] += 1
            try:
                if callable(read_object):
                    data = read_object(obj)
                else:
                    data = object
                if not isinstance(data, (dict, list)):
                    counts['object_read_invalid'] += 1
                    invalid_objects.append(obj)
                    continue
                if not isinstance(data, list):
                    data = [data]
            except Exception as e:
                invalid_objects.append(str(obj) + ": " + str(e))
                counts['objects_error'] += 1
                continue

            for index, d in enumerate(data):
                try:
                    counts['records_total'] += 1
                    if ((not isinstance(d, dict)) or (len(d) == 0)):
                        logger.error("Empty or invalid data",
                                     extra={
                                         'data': str(obj) + "\n" + str(d)[:100]
                                     })
                        counts['records_error_invalid'] += 1
                        continue

                    # Compute various feature sets for each patient
                    for detail in extractors:
                        try:
                            extractor = detail['extractor']
                            name       = detail['name']
                            
                            # Process one record...
                            features = extractor.one_record(d)
                            
                            # Skip if no features are being generated
                            if features is None:
                                continue

                            if isinstance(features, dict):
                                features = [features]
                            featuresets[name].extend(features)
                        except:
                            counts[f'extractor_{name}_exception'] += 1
                            if counts[f'extractor_{name}_exception'] == 1:
                                logger.exception(f"Unable to process:{name}",
                                                 extra={
                                                     'data': str(d)[:200]
                                                 })
                except:
                    # Handle exceptions in individual records
                    counts['records_error_exception'] += 1
                    logger.exception(f"Error in processing {index}",
                                         extra={
                                             'data': str(obj) + "\n" + str(d)[:200]
                                         })

                counts['objects_valid'] += 1

            # Cleanup
            try:
                del data
            except:
                pass

        except:
            # Handle exceptions in individual records
            counts['objects_error_exception'] += 1
            logger.exception(f"Error in processing object",
                             extra={
                                 'data': f"{obj}\n" + json.dumps(counts, indent=4)
                             })


    logger.debug("Completed reading objects",
                 extra={
                     'data': json.dumps(counts, indent=4) + "\nInvalid Objects:\n" + "\n".join(invalid_objects)
                 })

    # Now collect all features of all patient
    counts = defaultdict(int)
    computed = {}
    for detail in extractors:

        name = detail['name']
        extractor = detail['extractor']

        if ((name not in featuresets) or
            (featuresets[name] is None)):
            logger.warning(f"Features missing: {name}",
                           extra={
                           })
            continue

        # Collect all the features into a dataframe..
        df = extractor.collate(featuresets[name])

        # Clean the dataframe generated.
        df = extractor.clean(df)

        computed[name] = df
        counts[name] = df.shape[0]

    logger.debug("Completed collating",
                 extra={
                     'data': "records: " + json.dumps(counts, indent=4)
                 })
    
    # Now we have individual dataframes. May be the extractor
    # wants to compute some more.
    final = {}
    counts = defaultdict(int)
    for detail in extractors:
        name = detail['name']
        extractor = detail['extractor']
        df = computed.get(name, None)
        df = extractor.finalize(df, computed)
        if df is None:
            logger.error(f"{name}: Invalid result",
                         extra={
                         })
            continue
        final[name] = df
        counts[name] = df.shape[0]

    logger.debug("Completed finalization",
                 extra={
                     'data': "records: " + json.dumps(counts, indent=4)
                 })

    # Now document the outputs generated...
    for name, df in final.items():
        logger.debug(f"Featureset: {name}_features",
                     extra={
                         "data": extractor.document(name, df)
                     })

    return final



"""
Set of predefined tasks
"""

import os
import pytz
import gzip
import re
import json
import datetime
import requests
import subprocess
import platform
import logging
from enrichsdk.tasks import Task

logger = logging.getLogger("app")

__all__ = ['BackupCoreTask', 'SyncLocalCoreTask']

class CloudMixin(object):

    def get_params(self):

        # resolve the destination
        now = self.config.now()
        params = {
            'node': platform.node(),
            'date': now.strftime("%Y-%m-%d"),
            'datetime': now.strftime("%Y%m%d-%H%M%S")
        }
        return params

    def validate_args_cloud(self, what, state):

        dirspecs = self.dirspecs
        args = self.args

        fail = False
        msg = ""

        if "aws" not in args and "gcp" not in args:
            fail = True
            msg += "Both 'aws' or 'gcp' are missing. One is required\n"

        if "aws" in args:
            for v in ['credentials', 'bucket']:
                if v not in args['aws']:
                    fail = True
                    msg += "Element '{}' is missing in aws config\n".format(v)

            try:
                cred = self.get_credentials(args['aws']['credentials'])

                if (('access_key' not in cred) or ('secret_key' not in cred)):
                    fail = True
                    msg += "Invalid S3 credentials. Missing access_key or secret_key\n"

            except Exception as e:
                fail = True
                msg += str(e) + "\n"

        elif 'gcp' in args:
            for v in ['boto', 'bucket']:
                if v not in args['gcp']:
                    fail = True
                    msg += "Element '{}' is missing in gcp config\n".format(v)


        # => Now check the backup/sync specification
        if ((dirspecs not in args) or (not isinstance(args[dirspecs], list))):
            fail = True
            msg += "Element '{}' is missing or invalid (not a list)\n".format(dirspecs)
        else:
            for d in args[dirspecs]:
                if not isinstance(d, dict):
                    fail = True
                    msg += "Specification is not a dict: {}".format(d)
                    continue
                if 'src' not in d or 'dst' not in d:
                    fail = True
                    msg += "Specification should have src and dst"

        return fail, msg

    def validate_args(self, what, state):
        """
        Validate args.

        Looks like 'aws' and 'backupdirs' specification

        Example::

            "args": {
                 "backupdirs": [
                    {
                        "enable": true,
                        "name": "Logs",
                        "src": "%(enrich_root)s/logs/",
                        "dst": "backup/logs/"
                     },
                                ...
                  ],
                  "aws": {
                      "credentials": "enrich-acme",
                      "bucket": "client-enrich"
                  }
            }

        """
        args = self.args

        # Check
        fail, msg = self.validate_args_cloud(what, state)

        if fail:
            logger.error("Invalid configuration",
                         extra=self.config.get_extra({
                             'transform': self.name,
                             'data': msg

                     }))
            raise Exception("Invalid configuration")

    def get_env(self):
        """
        Get environment for the cloud provider
        """

        args = self.args
        env = dict(os.environ)
        if 'aws' in args:
            cred = self.get_credentials(args['aws']['credentials'])
            if cred is not None:
                if (('access_key' in cred) and (isinstance(cred['access_key'], str))):
                    env['AWS_ACCESS_KEY_ID'] = cred['access_key']
                if (('secret_key' in cred) and (isinstance(cred['secret_key'], str))):
                    env['AWS_SECRET_ACCESS_KEY'] = cred['secret_key']
        else:
            env['BOTO_CONFIG'] = self.config.get_file(args['gcp']['boto'])

        env['TZ'] = args.get('timezone', 'Asia/Kolkata')
        return env

    def get_command(self, d):
        """
        Get the command that must be run
        """

        # To or from cloud
        direction = self.direction

        # Extra params..
        params = self.get_params()

        if 'aws' in self.args:
            prefix = "s3"
            bucket = self.args['aws']['bucket']
        elif 'gcp' in self.args:
            prefix = "gs"
            bucket = self.args['gcp']['bucket']
        else:
            raise Exception("Unsupported source/destination")

        # Resolve the source and destination
        if direction == 'tocloud':
            src = self.config.get_file(d['src'])
            dst = d['dst'] % params
            dst = "{}://{}/{}".format(prefix, bucket, dst)
        else:
            dst = os.path.abspath(self.config.get_file(d['dst']))
            src = d['src'] % params
            src = "{}://{}/{}".format(prefix, bucket, src)

        if 'aws' in self.args:

            # aws s3 sync s3://mybucket s3://mybucket2
            if direction == 'tocloud' and os.path.isfile(src):
                awscmd = ['aws', 's3', 'cp', '--quiet']
            else:
                awscmd = ['aws', 's3', 'sync', '--quiet']

            if self.config.dryrun:
                awscmd.append("--dryrun")

            awscmd.extend([src, dst])
            return awscmd

        if 'gcp' in self.args:

            # gsutil sync gs://mybucket gs://mybucket2
            if direction == 'tocloud' and os.path.isfile(src):
                gscmd = ['gsutil', 'cp']
            else:
                gscmd = ['gsutil', '-m', '-q', 'rsync', '-r']

            if self.config.dryrun:
                gscmd.append("-n")

            # Add source and destination
            gscmd.extend([src, dst])



            return gscmd

        raise Exception("Unsupported source/destination")

    # => Now run the command
    def run_sync(self, state):
        """
        Run the backup task
        """

        env = self.get_env()

        dirspecs = self.args[self.dirspecs]
        for d in dirspecs:
            enable = d.get('enable', True)
            if not enable:
                continue

            name = d['name']

            # => What should I be running?
            cmd = self.get_command(d)

            # Now run the command
            p = subprocess.Popen(cmd,
                                 env=env,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            out, err = p.communicate()
            out = str(out, 'utf-8')
            err = str(err, 'utf-8')

            msg = """\
Cmd: {}
Error
-----
{}

Output:
-------
{}"""
            msg = msg.format(" ".join(cmd),
                             err,
                             out)

            # Log cloud-specific information
            if "BOTO_CONFIG" in env:
                msg += "Boto Config: {}".format(env['BOTO_CONFIG'])

            logger.debug("Backing up {}".format(name),
                         extra=self.config.get_extra({
                             'transform': self.name,
                             'data': msg
                         }))

        logger.debug("Sync'd all dirs",
                     extra=self.config.get_extra({
                         'transform': self.name,
                     }))

    def run(self, state):
        return self.run_sync(state)

class BackupCoreTask(CloudMixin, Task):
    """
    Backsup code and data as required
    """
    NAME = "BackupCoreTask"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "BackupCoreTask"
        self.dirspecs = 'backupdirs'
        self.direction = "tocloud"


class SyncLocalCoreTask(CloudMixin, Task):
    """
    Syncs a local directory with  content of a s3 bucket
    """
    NAME = "SyncLocalCoreTask"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "SyncLocalCoreTask"
        self.dirspecs = 'localdirs'
        self.direction = "fromcloud"




"""
Messaging
---------

Low-level messaging interface to connect to various
messaging backends such as Kafka and processing pipelines
such as spark streaming.

"""
import json
import logging
import traceback

# pykafka library. To be phased out..
from pykafka import KafkaClient

# kafka-python
from kafka import KafkaConsumer, KafkaProducer
from kafka.admin import KafkaAdminClient, NewTopic

logger = logging.getLogger("kafka")
logger.setLevel(logging.CRITICAL)

class MessagingBase(object):
    """
    Base class to handle the interfacing with various messaging
    backends.

    """
    def __init__(self, config, *args, **kwargs):
        """
        Initialize the messaging object

        Args:
            config (dict): Configuration
        """
        self.config = config
        if not isinstance(config, dict):
            raise Exception("Invalid config type. Expecting dict")

    def connect(self):
        """
        Initialize the messaging backend
        """
        pass

    def topcs(self):
        """
        List available topics
        """
        return []

    def produce(self, topic, callback, params):
        """
        Generate and post records into the stream

        Args:
            topic (str): Kafka topic to produce
            callback (method): method to call to generate stream input
            params (dict): Parameters to pass to callback

        Returns:
            None
        """
        raise Exception("Not implemented")

    def consume(self, topic, callback, params):
        """
        Generate and post records into the stream

        Args:
            topic (str): Kafka topic to consume
            callback (method): method to call to consume stream input
            params (dict): Parameters to pass to callback

        Returns:
            None
        """
        raise Exception("Not implemented")

    def stop(self):
        pass

class PyKafkaMessaging(MessagingBase):
    """

    Implementation of an interface to PyKafka

    Is being deprecated in favor of kafka-python
    implementation.

    Example::

        mg = PyKafkaMessaging({
            "hosts": ["127.0.0.1:9092", "127.0.0.1:9093"],
            "consume_params": {
                "consumer_group": 'testgroup',
                "auto_commit_enable": True,
                "zookeeper_connect": 'localhost:2181'
            })
        mg.connect()
        def handler(params, data):
            ....
        mg.consume('orders', handler, {'region': 'south'})

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'hosts' not in self.config:
            raise Exception("Kafka hosts must be specified")

    def connect(self):
        hosts = self.config.get('hosts')
        hosts = ",".join(hosts)
        self.client = KafkaClient(hosts=hosts)

    def produce(self, topic, callback, params):
        topic = self.client.topics[topic]
        producer_params = self.config.get('produce_params',{})
        with topic.get_sync_producer(**producer_params) as producer:
            for data in callback(params):
                producer.produce(json.dumps(data).encode('utf-8'))



    def consume(self, topic, callback, params):
        topic = self.client.topics[topic]
        consume_params = self.config.get('consume_params',{})
        balanced_consumer = topic.get_balanced_consumer(**consume_params)
        try:
            for message in balanced_consumer:
                if message is not None:
                    try:
                        callback(params, message)
                    except:
                        traceback.print_exc()
        finally:
            balanced_consumer.stop()

class KafkaMessaging(MessagingBase):

    def topics(self):
        consumer = KafkaConsumer(**self.config)
        available = consumer.topics()
        consumer.close()
        return available

    def create_topic(self, topic, partitions=1,replication_factor=1):
        admin_client = KafkaAdminClient(**self.config)

        topic_list = []
        topic_list.append(NewTopic(name=topic, num_partitions=1, replication_factor=1))
        admin_client.create_topics(new_topics=topic_list, validate_only=False)


    def produce(self, topic, callback, params={}):

        producer = KafkaProducer(value_serializer=lambda v: json.dumps(v).encode('utf-8'),
                             **self.config)

        try:
            if callable(callback):
                for data in callback(params):
                    producer.send(topic, data)
            else:
                data = callback
                producer.send(topic,data)
        finally:
            producer.close()


    def consume(self, topic, callback, params):

        consumer = KafkaConsumer(topic,
                                 value_deserializer=lambda v: json.loads(v.decode('utf-8')),
                                 **self.config)

        try:
            for msg in consumer:
                try:
                    callback(params, msg)
                except:
                    traceback.print_exc()
        finally:
            consumer.close()



from enrichsdk.lib.customer import find_usecase
from .models import *
from .forms import *

def get_spec():

    return {
        'name': 'Catalog',
        'description': "Catalog",
        'usecase': find_usecase(__file__),
        'models': {
            'catalog': Catalog,
            'attachment': Attachment,
            'datasource': DataSource,
            'column': Column,
            'role': Role,
            'visibilitymap': VisibilityMap
        },
        'forms': {
            'catalog': CatalogForm,
            'attachment': AttachmentForm,
            'datasource': DataSourceForm,
            'column': ColumnForm,
            'role': RoleForm,
            'roleselect': RoleSelectForm
        }
    }



from functools import partial
import pandas as pd

from .base import *
from .exceptions import *

class TableColumnsExistExpectations(ExpectationBase):
    """
    Check whether table has required columns

    Configuration has to specify a list of columns::

           {
             'expectation': 'table_columns_exist',
             'params': {
                 'columns': ['alpha', 'beta']
                 }
             }
           }

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "table_columns_exist"
        self.description = "Check whether columns exist (in any order)"

    def validate_args(self, config):
        super().validate_args(config)

        # Add extra
        self.args_validator_attrs_list(config, ['columns'])


    def generate(self, df):
        """
        Not implemented
        """
        return {
            'expectation': self.name,
            'description': self.description,
            'params': {
                'columns': list(df.columns)
            }
        }

    def validate(self, df, config):
        """
        Check if the specified columns are present in the dataframe
        """

        result = ExpectationResultBase()
        update = partial(result.add_result,
                         expectation=self.name,
                         description=self.description)

        if not isinstance(df, pd.DataFrame):
            update(passed=False,
                   extra={
                       'reason': 'Not a dataframe'
                   })
            return result


        columns = config['params']['columns']
        if isinstance(columns, str):
            columns = [columns]

        missing = [c for c in columns if c not in df.columns]
        if len(missing) > 0:
            update(passed=False,
                   extra={
                       'reason': 'Missing columns: {}'.format(",".join(missing))
                   })
            return result

        update(passed=True)
        return result

class TableColumnsPositionExpectations(ExpectationBase):
    """
    Check whether table has the right columns in right positions.

    Configuration has to specify the columns and their corresponding 
    positions in the test dataframe::

           {
             'expectation': 'table_columns_exist_with_position',
             'params': {
                 'columns': {
                    'alpha': 0,
                    'beta': 2
                 }
             }
           }

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "table_columns_exist_with_position"
        self.description = "Check whether columns exist (in particular order)"

    def validate_args(self, config):
        super().validate_args(config)

        # Add extra
        self.args_validator_attrs_list(config, ['columns'])

    def generate(self, df):
        """
        Not implementated yet.
        """
        columns = df.columns
        columns = { col: position for col, position in enumerate(columns)}
        return {
            'expectation': self.name,
            'description': self.description,
            'params': {
                'columns': columns
            }
        }

    def validate(self, df, config):
        """
        Validate the names and positions of columns
        """

        result = ExpectationResultBase()
        update = partial(result.add_result,
                         expectation=self.name,
                         description=self.description)

        if not isinstance(df, pd.DataFrame):
            update(passed=False,
                   extra={
                       'reason': 'Not a dataframe'
                   })
            return result

        columns = config['params']['columns']
        if not isinstance(columns, dict):
            raise InvalidConfigurationExpectation("Columns should be a dict")

        invalid = []
        existing = list(df.columns)
        for col, position in columns.items():
            if ((not isinstance(position, int)) or
                (position >= len(existing)) or
                (existing[position] != col)):
                invalid.append(col)

        if len(invalid) > 0:
            update(passed=False,
                   extra={
                       'reason': 'Missing/Wrong position columns: {}'.format(",".join(invalid))
                   })
            return result

        update(passed=True)
        return result




import traceback  

# Disable logging of dependent modules
import logging
logging.getLogger('urllib3').setLevel(logging.CRITICAL)
logging.getLogger('boto').setLevel(logging.CRITICAL)
logging.getLogger('boto3').setLevel(logging.CRITICAL)
logging.getLogger('botocore').setLevel(logging.CRITICAL)
logging.getLogger('nose').setLevel(logging.CRITICAL)
logging.getLogger("fsspec").setLevel(logging.WARNING)
logging.getLogger('s3fs').setLevel(logging.CRITICAL)
logging.getLogger('elasticsearch').setLevel(logging.CRITICAL)
logging.getLogger('elasticsearch.trace').setLevel(logging.CRITICAL)
logging.getLogger('numba.core.byteflow').setLevel(logging.CRITICAL)
logging.getLogger('numba.core.ssa').setLevel(logging.CRITICAL)
logging.getLogger('numba.core.interpreter').setLevel(logging.CRITICAL)

from .node import * 
from .sdk import * 
from .render import * 
from .mixins import * 
from .run import *
from .patch import *



import os 
import subprocess
import uuid
from hashlib import sha256, md5

def get_checksum(path): 
    
    if not os.path.isfile(path) or not os.path.exists(path): 
        return None 

    checksum = subprocess.check_output('/usr/bin/sha256sum {}'.format(path),        
                               shell=True)
    #'ffd4d29ee4019150cc308e46004d2f17105899983e7fb4ca7c27ffe981fb82d4  deploy.json\n'
    checksum = checksum.decode('utf-8').strip() 
    checksum = checksum.split(" ")[0] 
    return checksum

def get_file_size(path): 
    
    if not os.path.isfile(path) or not os.path.exists(path): 
        return 0

    statinfo = os.stat(path) 
    return statinfo.st_size

def get_consistent_uuid(text):
    # https://stackoverflow.com/questions/41186818/how-to-generate-a-random-uuid-which-is-reproducible-with-a-seed-in-python

    for n in range(0, 10):
        m = md5()
        m.update((text + " " *n).encode('utf-8'))
        new_uuid = uuid.UUID(m.hexdigest())
        if str(new_uuid)[0] == '0':
            continue
        else:
            break

    assert str(new_uuid)[0] != '0'
    return new_uuid



import os
import sys
import json
import logging
from datetime import datetime
import platform
import distro
import getpass
import logging
from collections import OrderedDict
import logging

from . import log

logger = logging.getLogger('app')

class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        else:
            return super().default(obj)

#######################################################
# Config object to handle a number of helper functions
#######################################################
class Config(object):
    """
    Context object having a number of capabilities including
    log management, files and path management, credentials
    lookup etc.

    """
    def __init__(self, params, *args, **kwargs):
        """
        Initialize the config (context) object.

        Args:
          params (dict): Only two keys are supported. debug (bool) prints
                         extra information. It is useful for debugging loading.
                         settings is the path to the settings file.
        """
        super().__init__(*args, **kwargs)
        self.debug = params.get('debug', False)
        self.settings = params.get('settings', None)
        self.readonly = params.get('readonly', False)
        self.show_metadata = params.get('show_metadata', True)

        if self.settings is not None:
            self.settings = os.path.expanduser(self.settings)

        self.now = datetime.now()
        self.conf = {
            'credentials': {},
            'libraries': []
        }
        self.stats = {
            'platform': {
                'node': platform.node(),
                'os': platform.system(),
                'release': platform.release(),
                'processor': platform.processor(),
                'python': platform.python_version(),
                'distribution': distro.linux_distribution()
            },
            'user': getpass.getuser(),
        }

        self.state = self.get_default_metadata()
        self.state.update({
            'outputs': [],
            'notes': [],
            'message': ""
        })

    def get_default_metadata(self):
        """
        Get reusable metadata dict with some standard fields.
        """
        metadata = {
            "schema": "standalone:command",
            "version": "1.0",
            'timestamp': datetime.now().replace(microsecond=0).isoformat(),
            'command': {
                "name": "Unknown",
                'status': 'initial',
                "description": "",
                "host":  self.stats['platform']['node'],
                "stats": self.stats,
                "pid":  os.getpid(),
                'start_time': self.now,
                'end_time': self.now,
                "cmdline":  list(sys.argv),
            }
        }

        return metadata

    def configure(self, settings=None):
        """
        Configure the config object

        Args:
           settings (str): Path to the settings file

        """
        if settings is None:
            settings = self.settings

        if ((settings is None) or
            (not os.path.exists(settings))):
            # Quietly return. Do nothing..
            raise Exception("Settings path invalid or undefined: {}".format(settings))

        try:
            loaded_conf = json.load(open(settings))
        except:
            logger.exception("Invalid settings file")
            loaded_conf = None

        if loaded_conf is None:
            return

        # Materialize the paths...
        self.enrich_root = loaded_conf.get('enrich_root')
        self.enrich_root = os.path.expanduser(self.enrich_root)
        self.enrich_root = os.path.abspath(self.enrich_root)

        libraries = loaded_conf.get('libraries', [])
        libraries = [self.get_file(l) for l in libraries]
        self.conf['libraries'] = libraries

        # Insert the rest
        defaults = {
            "enrich_data": "%(enrich_root)s/data",
            "enrich_customers": "%(enrich_root)s/customers",
            "runid": "run-%Y%m%d-%H%M%S-%(user)s",
            "command_root": "%(enrich_data)s/_commands",
            "rundir": "%(command_root)s/%(hostname)s/%(runid)s",
            "outputdir": "%(rundir)s/output",
            "log": "%(rundir)s/log.json",
            "metadata": "%(rundir)s/metadata.json",
            'args': {}
        }

        for name in ['enrich_data', 'enrich_customers',
                     'runid', 'command_root',
                     'rundir', 'outputdir', 'log',
                     'metadata']:
            default = defaults[name]
            try:
                final = loaded_conf.get(name, default)
                final = self.get_file(final)
            except:
                error = "Could not resolve: {} ({})".format(name, final)
                raise Exception(error)

            setattr(self, name, final)
            self.state['command'][name] = final

        for name in ['args']:
            default = defaults[name]
            try:
                final = loaded_conf.get(name,default)
            except:
                error = "Could not resolve: {} ({})".format(name, final)
                raise Exception(error)

            setattr(self, name, final)

        # Set environment variables.....
        os.environ['ENRICH_ROOT'] = self.enrich_root
        os.environ['ENRICH_CUSTOMERS'] = self.enrich_customers
        os.environ['ENRICH_DATA'] = self.enrich_data

    def get_file(self, path):
        """
        Resolve the abstract specification into a full path

        Args:
           path (str): Path specification

        Example::

           %(enrich_data)/acme/Projects/commands

        This is resolved into `/home/ubuntu/enrich/data/acme/Projects/commands`

        """
        path = path.replace("%(",'%%(') # This will make the strftime handling safe
        path = self.now.strftime(path)
        path = path % {
            'enrich_root': self.enrich_root,
            'enrich_data': getattr(self, 'enrich_data', ''),
            'enrich_customers': getattr(self, 'enrich_customers', ''),
            'user': self.stats['user'],
            'username': self.stats['user'],
            'hostname': self.stats['platform']['node'],
            'node': self.stats['platform']['node'],
            'runid': getattr(self, 'runid', ''),
            'rundir': getattr(self, 'rundir', ''),
            'outputdir': getattr(self, 'outputdir', ''),

        }
        return path

    def get_relative_path(self, path,what='enrich_root'):

        if what == 'enrich_root':
            return os.path.relpath(path, self.enrich_root)
        elif what == 'enrich_data':
            return os.path.relpath(path, self.enrich_data)
        else:
            raise Exception("relative path should specify the root")

    def add_output(self, params):
        """
        Add file to the command output

        Args:
           params (dict): file parameters
              File parameters includes a name of the output, relative path
              to the file, and a description for the output

        """
        required = ['name', 'description', 'path']
        missing = [n for n in required if n not in params]
        if len(missing) > 0:
            raise Exception("Missing parameters for the output: {}".format(missing))

        if os.path.isabs(params['path']):
            raise Exception("Path specified for output {} should be relative".format(params['name']))

        for o in self.state['outputs']:
            if o['name'] == params['name']:
                raise Exception("An output already exists with the name: {}".format(params['name']))

        params['fullpath'] = self.get_file("%(outputdir)s/{}".format(params['path']))
        params['relpath'] = self.get_relative_path(params['fullpath'])

        # Now append
        self.state['outputs'].append(params)

    def get_output_handle(self, name, *args):
        """
        Get a file descriptor to the named output.

        Args:
           name (str): Name of the output.
              This output was previously added using `add_output` call.

        """
        for o in self.state['outputs']:
            if o['name'] == name:
                try:
                    os.makedirs(os.path.dirname(o['fullpath']))
                except:
                    pass
                return open(o['fullpath'], *args)

        raise Exception("Could not find output {}".format(name))

    def get_library_paths(self):
        return self.conf['libraries']


    def get_args(self, cmd):
        """
        Get default arguments to the command specified in the settings file

        Args:
           cmd (str): Name of the command

        """
        logger = logging.getLogger('app')
        if not hasattr(self, 'args'):
            if self.debug:
                logger.error("args not present")
            return {}

        if not isinstance(self.args, dict):
            if self.debug:
                logger.error("args not a dictionary")
            return {}

        return self.args.get(cmd, {})

    # Update the state
    def set_status(self, value):
        self.state['command']['status'] = value

    def set_message(self, value):
        self.state['message'] = value

    def dump_state(self):

        # Update file statistics
        for o in self.state['outputs']:
            path = o['fullpath']
            st = os.stat(path)
            o['size'] = st.st_size

        metadata = OrderedDict([
            ('schema', self.state['schema']),
            ('version', self.state['version']),
            ('timestamp', self.state['timestamp']),
            ('command', self.state['command']),
            ("message", self.state['message']),
            ("outputs", self.state['outputs']),
            ("notes", self.state['notes']),
            ("log", os.path.relpath(self.log, self.rundir)),
        ])

        if not self.readonly:
            with open(self.metadata, 'w') as fd:
                json.dump(metadata, fd, indent=4, cls=CustomEncoder)
        elif self.show_metadata:
            print(json.dumps(metadata, indent=4, cls=CustomEncoder), file=sys.stderr)

        if self.debug:
            logger.debug("Dumped execution metadata")

    # Logging start and end
    def start_run(self, cmd):

        # Override with parameters from settings
        if not self.readonly:
            args = self.get_args(cmd)
            self.readonly = args.get('readonly', self.readonly)
            self.show_metadata = args.get('show_metadata', self.show_metadata)

        # Setup logging
        if self.readonly:
            log.logging_start()
        else:
            log.logging_start(self.log)

        logger = logging.getLogger('app')
        if self.debug:
            logger.debug("Starting run")

        # Dump the state at the beginning
        self.dump_state()

    def end_run(self):

        self.state['command']['end_time'] = datetime.now()
        self.dump_state()

        # Dump status
        logger = logging.getLogger('app')

        if self.debug:
            logger.debug("Ending run")
        log.logging_end()





"""Module to set and implement `expectations` on dataframes. Note
that this is development mode and meant to be a preview.

As systems are getting automated, and processing more data everyday,
it is hard to keep track of correctness of the entire system. We use
expectations to make sure that decision modules such as ML/statistics
code is being correctly fed.

An expectation is a set of structural rules that datasets should
satisfy. It could be superficial such as names or order of columns. It
could go deeper and specify the statistical attributes of the input.

This module help implement a specification. An example is shown
below::

        expectations = [
            {
                'expectation': 'table_columns_exist',
                'params': {
                    'columns': ['Make', 'YEAR', 'Model', 'hello']
                }
            },
            {
                'expectation': 'table_columns_exist_with_position',
                'params': {
                    'columns': {
                        'Make': 1
                    }
                }
            }
        ]


Each expectation has a name, and arbitrary parameters. The first
expectation shown above will for example, check if the specified
columns exist in the dataframe being evaluated. The result will be as
shown below::

        {
            "expectation": "table_columns_exist_with_position",
            "passed": true,
            "description": "Check whether columns exist (in particular order)",
            "version": "v1",
            "timestamp": "2020-01-04T11:43:03+05:30",
            "frame": "v2scoring",
            "meta": {
                "node": "data.acmeinc.com",
                "pipeline": "AcmeLending",
                "transform": "ScoringModule",
                "runid": "scoring-20200104-114137",
                "start_time": "2020-01-04T11:41:37+05:30"
            }
        }


This is stored as part of the state and dumped in the metadata for
each run.

The expectations are used as follows::

        expectationsfile=os.path.join(thisdir, "expectations.py")
        checker = TransformExpectation(self,
                                       mode='validation',
                                       filename=expectationsfile)

        result = checker.validate(df,selector=selector)
        state.add_expectations(self, name, result)

"""

from .base import *
from .expectations import *
from .exceptions import *
from .transforms import *



"""
Module to facilitate access to Enrich Server API
and also show on the command line.

Visit <server>/api/v1.0/schema to understand the available
API calls.

"""

import json
import requests
import re
from collections import OrderedDict

from .draw import *

class Backend():
    """
    Class to access Enrich Server API
    """

    def __init__(self, config):
        """
        Initialize an access object

        Args:

           config (dict): Should have 'server' and 'key'.
        """

        self.config = config
        assert 'server' in config
        assert 'key' in config

    def get_api_url(self, path, use_prefix=True):

        path = path.replace("//","/")
        server = self.config['server']

        # Avoid double lines...
        if server[-1] == "/" and path[0] == "/":
            path = path[1:]

        if use_prefix:
            apiversion = self.config.get('version', 'v1.0')
            url = "{}/api/{}/{}".format(server, apiversion, path)
        else:
            url = "{}/{}".format(server, path)

        return url

    def call(self, url, params={}, method="get", data={}):
        """
        Call backend with some extra query parameters

        Args:

           path (str): API path suffix (after api/v1.0/)
           params (dict): Query parameters
        """

        method = str(method).lower()

        key    = self.config['key']
        headers={
            'Authorization': "Token {}".format(key),
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        try:
            if method == 'get':
                response = requests.get(url,
                                        verify=False, # ignore SSL checks
                                        headers=headers,
                                        params=params)
            elif method == 'post':
                response = requests.post(url,
                                        verify=False,
                                        headers=headers,
                                        params=params,
                                        json=data)
            else:
                raise Exception("Only get/post supported")

            if response.status_code not in [200, 400, 404]:
                return {
                    'status': 'error',
                    'message': response.content[:100].decode('utf-8')
                }
            return response.json(object_pairs_hook=OrderedDict)

        except requests.exceptions.ConnectionError as e:
            return {
                "status": "error",
                "message": "Connection refused"
            }
        except Exception as e:
            return {
                "status": "error",
                "message": str(e)
            }

    def status(self, offset=0, page=20):
        """
        Shows a list of tasks/pipelines that ran/are running

        Args:
           offset (int): Offset in the execution list
           page (int): Page in the execution list
        """

        url = self.get_api_url('status')
        result = self.call(url, {
            'offset': offset,
            'page': page
        })

        return result

    def usecases(self):
        """
        Show all the usecases available on server
        """
        url = self.get_api_url('usecases')
        result = self.call(url)
        return result

    def health(self):
        url = self.get_api_url('system/health')
        result = self.call(url)
        return result

    def pipelines(self, usecase):
        """
        Shows all pipelines within a given usecase

        Args:
           usecase (str): Name of the usecase
        """
        url = self.get_api_url('usecase/pipelines')
        result = self.call(url, {
            'usecase': usecase
        })
        return result

    def run_list(self, usecase, pipeline):
        """
        Shows all runs of a given pipeline

        Args:
           usecase (str): Name of the usecase
           pipeline (str): Name of the pipeline

        """
        url = self.get_api_url('usecase/pipeline/runs')
        result = self.call(url, {
            'usecase': usecase,
            'pipeline': pipeline
        })
        return result

    def run_detail(self, usecase, pipeline, runid):
        """
        Shows details of a given run of a pipeline

        Args:
           usecase (str): Name of the usecase
           pipeline (str): Name of the pipeline
           runid (str): Name of the run

        """
        url = self.get_api_url('usecase/pipeline/run/detail')
        result = self.call(url, {
            'usecase': usecase,
            'pipeline': pipeline,
            'runid': runid
        })
        return result

    def tasks(self, usecase):
        """
        Shows all tasks within a given usecase

        Args:
           usecase (str): Name of the usecase
        """
        url = self.get_api_url('usecase/tasks')
        result = self.call(url, {
            'usecase': usecase
        })
        return result

    def task_run_list(self, usecase, task):
        """
        Shows all runs of a given task

        Args:
           usecase (str): Name of the usecase
           task (str): Name of the task
        """
        url = self.get_api_url('usecase/task/runs')
        result = self.call(url, {
            'usecase': usecase,
            'task': task
        })
        return result

    def task_run_detail(self, usecase, task, runid):
        """
        Shows details of a run

        Args:
           usecase (str): Name of the usecase
           task (str): Name of the task
           runid (str): Name of a run
        """
        url = self.get_api_url('usecase/task/run/detail')
        result = self.call(url, {
            'usecase': usecase,
            'task': task,
            'runid': runid
        })
        return result




import os
import shutil
import logging

logger = logging.getLogger('app')

class OutputHandler(object):
    """
    How should output be handled? 
    """
    def __init__(self, transform, config):
        self.transform = transform
        self.config = config

    def process(self, tempfile):
        raise Exception("Not Implemented")

class FileOutputHandler(OutputHandler):
    """
    Config in this case is a simple path.
    """
    def exists(self, params):
        output = self.config
        output = self.transform.get_file(output, abspath=False, extra=params)
        return os.path.exists(output)

    def process(self, tempfile, params):

        output = self.config
        output = self.transform.get_file(output,
                                         abspath=False,
                                         extra=params)

        # Move
        try:
            os.makedirs(os.path.dirname(output))
        except:
            pass 
        shutil.move(tempfile, output)
        logger.debug("Handled output",
                     extra={
                         'transform': self.transform.name,
                         'data': "From: {}\nTo: {}".format(tempfile, output)
                     })

def get_mysql_config(cred):
    mapping = {
        "host": ['HOST', 'host', 'hostname'],
        "user": ['USER', 'user', 'username'],
        "port": ["port", "PORT"],
        'password': ['PASSWORD', 'PASSWD', 'password', 'passwd', 'pwd'],
        'db': ['db', 'database', 'NAME', 'name'],
        'options': ['OPTIONS', 'options']
    }

    config = {}
    for k, alternatives in mapping.items():
        for alt in alternatives:
            if alt in cred:
                config[k] = cred[alt]
                break

    options = config.get('options', {})
    config.update(options)
    return config



import os 
import re 
import json
from ..lib import renderlib

class Widgets(object):     
    """
    List of available widgets 
    """
    pass 

############################################
# Update the widget class with attributes. 
############################################
thisdir = os.path.dirname(__file__) 
templatedir = os.path.join(thisdir, '..', 'templates')
templatedir = os.path.abspath(templatedir) 
widgetdir = os.path.join(templatedir, 'widgets') 

def get_func(attrname, filename): 
    def dummyfunc():
        pass 
        
    content = open(os.path.join(widgetdir, filename)).read()
    try: 
        content = re.search(r"<!--(.*)-->", content).group(1)
    except:
        content = "Widget" 
        
    # Get variables...
    variables = renderlib.template_get_variables(widgetdir, 
                                                 attrname)
    
    if len(variables) > 0: 
        content += """

    Args:

"""
        for v in variables: 
            content += "        " + v + "(str): parameter\n" 

    content += """
    Returns:

       Rendered html 
"""
    dummyfunc.__doc__ = content 
    return dummyfunc 

for filename in os.listdir(widgetdir):     
    if not filename.endswith(".html"):
        continue

    attrname = filename.replace(".html","")
    attr = get_func(attrname, filename)
    setattr(Widgets, attrname, property(attr))



import os
import sys 
from enrichsdk.policy import * 

class PipelineRunResource(BaseResource): 
    """
    Resource that implements access to pipeline output
    """

    def __init__(self, run, *args, **kwargs): 
        super().__init__(*args, **kwargs) 
        self.run = run 
        
    def name(self): 
        return "PipelineRun::{}::{}".format(self.run.config.name, 
                                            self.run.name)
        
class SimpleFilterEngine(BasePolicyEngine): 
    """
    This applies the policy on a collection of dataframes 
    """
    NAME = "SimpleFilterPolicy" 

    def __init__(self, *args, **kwargs): 
        self.config = kwargs.pop('config') 
        super().__init__(*args, **kwargs) 
        self.name = "SingleApplicationPolicy" 
        self.filter_handlers = { 
            "by_column_value": self.filter_by_column_value, 
            "by_column_range": self.filter_by_column_range
        }

    def filter_by_column_value(self, principal, resource, params): 
        """
        Filter the data by value of a particular column. The 
        column and the value is specified in policy 

        Args: 
          principal (obj): Principal object 
          resource (obj): Resource object
          params (dict): Filter parameters
        """
        pass 

    def filter_by_column_range(self, principal, resource, params): 
        """
        Filter the data by the range of values of a particular
        column. The column and the value is specified in policy. 

        Args: 
          principal (obj): Principal object 
          resource (obj): Resource object
          params (dict): Filter parameters
        """
        pass 




from sqllineage.runner import LineageRunner

from .sample import *

def get_lineage_of_query(engine, sql):
    """
    Given an engine and an SQL expression, find the input
    and output tables. Replace <default> with database 
    name from the engine
    """
    dependencies = []
    result = LineageRunner(sql, verbose=True)

    database = engine.url.database

    # Get input tables..
    if len(result.source_tables) > 0:
        tables =[str(t).replace("<default>",database) for t in result.source_tables]

        tables = {
            t.split(".")[-1] : {
                'text': t.replace(".", " "),
                'table': t
            } for t in tables
        }
        dependencies.append({
            "type": "db",
            "nature": "input",
            "objects": tables,
        })

    # Get output dependencies
    if len(result.target_tables) > 0:
        tables =[str(t).replace("<default>",database) for t in result.target_tables]
        tables = {
            t.split(".")[-1] : {
                'text': t.replace(".", " "),
                'table': t
            } for t in tables
        }

        dependencies.append({
            "type": "db",
            "nature": "output",
            "objects": tables
        })

    return dependencies




import os 
import json 
import traceback 

from flask import Flask, render_template, request 
from flask_multistatic import MultiStaticFlask

app = MultiStaticFlask(__name__)

@app.route("/")
def main():

    print(request.args) 
    # First get the arguments if any 
    params = {} 
    for name in request.args: 
        valuelist = request.args.getlist(name) 
        params[name] = valuelist 

    try: 
        skin = app.config['skin'] 
        data = skin.render(params)
    except: 
        traceback.print_exc() 
        data = [{
            'target': 'result', 
            'rendered': "<p class='lead'>Error while rendering</p>" 
            }]

    return render_template('index.html', data=data)

def shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()

@app.route('/shutdown')
def shutdown():
    shutdown_server()
    return('Server shutting down...')

# Custom static data
@app.route('/app/static/<path:filename>')
def custom_static(filename):
    for p in app.config['static_paths']: 
        fullpath = os.path.join(p, filename)
        if os.path.exists(fullpath): 
            return send_from_directory(p, filename)

            
def create_app(conf): 

    # => Set the static folder...
    static_paths = conf.pop('static_paths') 
    static_paths.insert(0, os.path.join(app.root_path, 'static'))
    app.static_folder = static_paths 

    app.config.update(conf) 
    return app

if __name__ == "__main__":
    app.run(debug=True)



from django.conf.urls import url, include
from enrichapp.dashboard.control.urls import control_urlpatterns
from enrichapp.dashboard.audit.urls import audit_urlpatterns
from enrichapp.dashboard.catalog.urls import catalog_urlpatterns
from enrichapp.dashboard.persona.urls import persona_urlpatterns
from enrichapp.dashboard.filerenderer.urls import filerenderer_urlpatterns
from enrichapp.dashboard.marketplace.urls import marketplace_urlpatterns, get_marketplace_urls

from . import views, control, catalog

app_name = "APPNAME"

# => Base
urlpatterns = [
    url('^[/]?$', views.index, name="index"),

]

urlpatterns += [
    url(r'^control/', include((control_urlpatterns,"control"),
                              namespace="control"),
        {
            'spec': control.get_spec()
        })
]

urlpatterns += [
    url(r'^catalog/', include((catalog_urlpatterns,'catalog'),
                              namespace="catalog"),
        {
            'spec': catalog.get_spec()
        })
]


# Enable it
#from . import persona
#searchspec = persona.get_spec()
#urlpatterns += [
#    url(r'^persona/', include((persona_urlpatterns, "persona"),
#                               namespace="persona"),
#        {
#            'spec': searchspec
#        })
#]



import os
import sys
import json
import s3fs

def s3_sample_crawl(root, depth=3, ext="json"):

    s3 = s3fs.S3FileSystem(anon=False)

    prefix = root.replace("s3://","")
    print("Crawling")
    for subroot, dirs, files in s3.walk(root,maxdepth=depth):
        relpath = os.path.relpath(subroot, start=prefix)
        if len(files) > 0:
            print(relpath)
        for f in files[:3]:
            basename = os.path.basename(f)
            print("  ", basename, ":", os.path.join("s3://", subroot, f))

def s3_sample_cat(path):

    s3 = s3fs.S3FileSystem(anon=False)            
    content = s3.cat(path)
    content = content.decode('utf-8')
    print(content)
    
def sample_cat(path):

    s3 = s3fs.S3FileSystem(anon=False)

    if path.startswith("s3://"):
        if 'AWS_PROFILE' not in os.environ:
            print("Please define AWS_PROFILE")
            return
        return s3_sample_cat(path)
    else:
        print("Unknown root. Only s3://<path> is supported for now")
    

def sample_crawl(root, ext):

    if root.startswith("s3://"):
        if 'AWS_PROFILE' not in os.environ:
            print("Please define AWS_PROFILE")
            return
        return s3_sample_crawl(root=root, ext=ext)
    else:
        print("Unknown root. Only s3://<path> is supported for now")



import copy 
import json 
import logging 

logger = logging.getLogger('app') 

__all__ = ['Task']

class Task(object): 
    """
    This is the task base class. This is 'executed' by 
    a 'runner' 
    """
    NAME = "BaseTask" 

    def __init__(self, *args, **kwargs): 
        self.config = kwargs.pop('config') 
        self.name = "BaseTask" 
        self.description = "Baseclass of tasks" 
        self.version = "1.0" 
        self.testdata = {
            'data': {} 
        } 
        
        self.supported_extra_args = []
        """Extra arguments that can be passed on the commandline to 
        the module. 

        self.extra_args = [
               {
                  "name": "jobid",
                  "description": "Export JobID",
                  "type": "str",
                  "required": True,
                  "default": "22832",
               }
        ]
        """

    #################################################
    # Configuration the Module 
    #################################################
    def preload_clean_args(self, args):
        """
        Clean the arguments before using them 
        
        Parameters
        ----------
        args: Args specified in the config file         
        """


        if not self.config.enable_extra_args: 
            return args
        
        # Override the defaults with the parameters 
        # Three sources for a given args attribute in the order of
        # priority
        # 
        # 1. command line
        # 2. Configuration
        # 3. Default

        readonly = self.config.readonly 
        cmdargs = self.config.get_cmdline_args()         
        supported_extra_args = self.supported_extra_args

        # Collect the final values for all the defaults 
        final = {} 
        for a in supported_extra_args:
            name = a['name']
            cmdlabel = "{}:{}".format(self.name, name) 
            default = a['default']
            required = a['required'] 
            if required and not readonly: 
                # Required and specified on the command line...
                if cmdlabel not in cmdargs:
                    logger.error("Invalid configuration",
                                 extra={
                                     'transform': self.name,
                                     'data': "Variable {} for task {} is required but missing in command line".format(name, self.name)
                                 })                    
                    raise Exception("Invalid configuration") 
                final[name] = cmdargs[cmdlabel]
            elif cmdlabel in cmdargs:
                # Optional but specified on the command line 
                final[name] = cmdargs[cmdlabel] 
            elif name in args:
                # Priority 2
                final[name] = args[name]
            else:
                # Priority 3 
                final[name] = default 

        # => Log what has happened...
        msg = ""
        for k,v in final.items():
            msg += "{}: {}\n".format(k,v) 
        logger.debug("Configuration has been overridden",
                     extra={
                         'transform': self.name,
                         'data': msg
                     })
        
        args.update(final) 
            
                
        return args

    def preload_validate_self(self): 
        """
        Check whether definition of the task is fine 

        """
        supported_extra_args = self.supported_extra_args
        if not isinstance(supported_extra_args, list):
            logger.error("Invalid configuration. Expected {}, actual {}".format('list', str(type(supported_extra_args))),
                         EXTRA=self.config.get_extra({
                             'transform': self.name,
                             
                         }))
            raise Exception("Supported extra argument should be a list of dicts")

        for i, a in enumerate(supported_extra_args): 
            if not isinstance(a, dict):
                logger.error("Invalid configuration. Entry {} of supported_extra_args: Expected {}, actual {}".format(i, 'dict', str(type(a))),
                             EXTRA=self.config.get_extra({
                                 'transform': self.name,
                                 
                         }))
                raise Exception("Supported extra argument should be a list of dicts")

            required = ['name', 'description', 'default', 'required']
            missing = [x for x in required if x not in a] 
            if len(missing) > 0:
                logger.error("Invalid configuration. Entry {} of supported_extra_args: Missing attributes: {}".format(i, ", ".join(missing)),
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': str(a) 
                                 
                         }))
                raise Exception("Missing attributes in one of the supported extra args")

    def get_default_metadata(self, state):
        """
        Get reuse metadata dict with some standard fields.
        """
        metadata = {
            "schema": "standalone:task",
            "version": "1.0", 
            'timestamp': datetime.now().replace(microsecond=0).isoformat(),             
            'task': {
                "usecase":   self.config.usecase["org"]["name"],
                # Backward compatability
                "customer":   self.config.usecase["org"]["name"], 
                "name":   state.name,
                "description":   self.config.description,
                "host":  state.stats['platform']['name'],
                "runid":   state.runid,
                "pid":  state.pid,
                "start_time":   state.start_time.replace(microsecond=0).isoformat(),
                "end_time":   state.end_time.replace(microsecond=0).isoformat(),
                "versionmap":  self.config.get_versionmap(),
                "stats":   state.stats,
                "cmdline":  list(sys.argv),
            }
        }
        
        return metadata
    
    def preload_validate_conf(self, conf):
        """
        Check whether this configuration is even valid?

        Parameters
        ----------
        conf: Module configuration specified in the config file
        """

        if not (isinstance(conf, dict) and len(conf) > 0): 
            logger.error("Conf is not a valid dictionary",
                         extra=self.config.get_extra({
                             'transform': self.name
                         }))
            raise Exception("Conf is not a valid dictionary")

        if self.version != conf.get('version', "1.0"):
            logger.error("Version mismatch",
                         extra=self.config.get_extra({
                             'transform': self.name
                         }))
            raise Exception("Version mismatch between config and module")

    def get_supported_extra_args(self): 
        """
        What extra command line arguments are supported 

        Returns
        -------
        list: List of arg specifications 
 
        see supported_extra_args 
        """
        return copy.copy(self.supported_extra_args)
    
    def configure(self, spec): 
        """
        Configure this task  
        """

        # => Check whether internal configuration is valid 
        self.preload_validate_self()

        # => Check whether passed configuration is fine
        self.preload_validate_conf(spec) 

        # Make a copy of the spec 
        self.spec = copy.deepcopy(spec) 

        # Override attributes from the spec 
        for attr in ['name', 'output_version', 'test', 'enable']: 
            if ((attr in spec) and hasattr(self, attr)): 
                setattr(self, attr, spec[attr])

        # Clean and load args 
        args = spec.get('args', {}) 
        self.args = self.preload_clean_args(args) 

        
    #=> Validation code 
    def validate_conf(self, what, state): 
        pass 

    def validate_args(self, what, state): 
        pass 

    def validate_results(self, what, state): 
        pass 

    def validate_testdata(self, what, state): 
        
        if not hasattr(self, 'testdata'): 
            raise Exception("testdata element is missing") 
            
        testdata = self.testdata 
        if (not isinstance(testdata, dict) or 
            ('conf' not in testdata) or 
            ('data' not in testdata)): 
            raise Exception("testdata should be dictionary with conf and data elements") 

    def validate(self, what, state): 
        """
        Validate various aspects of the task state,
        configuration, and data. Dont override this function. Override
        specific functions such as `validate_args`. 

        Args:
            what (str): What should be validated? args, conf, results etc.

        Returns:

             Nothing

        Raises:
             Exception ("Validation error") 

        """

        func = getattr(self, 'validate_' + what, None)
        if hasattr(func, '__call__'):
            return func(what, state)
        else:
            logger.error("Cannot find function to validate {}".format(what),
                         extra=self.config.get_extra({
                             'transform': self.name
                         }))
            raise Exception("Validation error")


    def initialize(self): 
        """
        Initialize the task by connecting to database etc.
        """
        pass 

    def get_credentials(self, name): 
        """
        Helper function to access siteconf for credentials 
        """
        if not hasattr(self.config, 'siteconf'): 
            raise Exception("No siteconf") 

        siteconf = self.config.siteconf 
        if not isinstance(siteconf, dict): 
            raise Exception("Invalid siteconf format") 
            
        try: 
            credentials = siteconf['credentials'][name]
        except: 
            logger.exception("Cannot acquire credentials", 
                             extra=self.config.get_extra({
                                 'transform': self.name 
                             }))
            raise Exception("Missing/invalid credentials") 
        
        return credentials 

    def run(self, *args, **kwargs): 
        """
        Main execution func. Override this 
        """
        raise Exception("Should be implemented by the derived class")


if __name__ == "__main__": 
    print(vars())
    subclasses = vars()['Task'].__subclasses__() 
    for cls in subclasses: 
        print(cls.NAME) 



from __future__ import unicode_literals
from django.db import models
import jsonfield
from enrichapp.dashboard.catalog.modellib import *
from enrichapp.dashboard.marketplace.modellib import *

# Rename/add/delete
from enrichapp.dashboard.persona.modellib import *
class ComplexSearchRecord(SearchRecordBase):
    pass

##################################################
# Regular Models
##################################################
class Catalog(CatalogBase):
    pass

class Attachment(AttachmentBase):
    pass

class DataSource(DataSourceBase):
    pass

class Column(ColumnBase):
    pass

class Role(RoleBase):
    pass

class VisibilityMap(VisibilityMapBase):
    pass


class Comment(CommentBase):
    pass

class FeatureRequest(FeatureRequestBase):
    pass

class Experiment(ExperimentBase):
    pass

class ExperimentRun(ExperimentRunBase):
    pass

class Feature(FeatureBase):
    pass



import sys
import inspect
import copy
from ..lib.customer import find_usecase
from django.apps import AppConfig

class EnrichAppConfig(AppConfig):

    filename = None    
    enable = False
    multiple = False
    composition = False

    def get_usecase(self):

        if self.filename is None:
            return {}

        usecase = find_usecase(self.filename)
        self.usecase = copy.deepcopy(usecase)
        return self.usecase

    def is_composition(self):
        return self.composition

    def get_name(self):
        return self.name

    def get_verbose_name(self):
        return self.verbose_name

    def get_description(self):
        return self.description

    def is_enabled(self):
        return self.enable




"""
Services to allow enrichsdk to be used in notebooks. It
provides:

1. Security - access to obfuscated credentials
2. Resource - limit resource usage
3. Indexing - search notebooks
4. Metadata - Generate metadata to be included in the output files

"""
import os
import sys
import platform
import distro
import resource
import psutil
import nbformat
import glob
import json
import warnings
from datetime import datetime, date
from shutil import copy
import mimetypes
import logging
import pandas as pd
from collections import OrderedDict

from .utils import *

logger = logging.getLogger('app')

class Notebook(object):
    """
    This class allows one to read and search in notebooks
    """
    def __init__(self, data_root=None):
        if data_root is not None:
            if not os.path.exists(data_root):
                raise Exception("Invalid or missing path or file: {}".format(data_root))
            else:
                self.data_root = data_root
                self.output_dir = os.path.join(data_root,"output")
        else:
            if 'JUPYTERLAB_HOME' not in os.environ:
                warnings.warn('Notebook using JUPYTERLAB_HOME when being executed outside jupyter environment. Either use data_root, define the variable or dont use this class')
                raise Exception("JUPYTERLAB_HOME not defined")

            self.data_root = os.path.join(os.environ['JUPYTERLAB_HOME'],"data")
            self.output_dir = os.path.join(os.environ['JUPYTERLAB_HOME'],"data","output")


    def set_resource_limits(self, params={}):
        """
        Set resource limits for the notebook. This applies
        only to this notebook.

        Parameters
        ----------
        params: Only one member (memory) is accessed. It is fraction
                of available memory to be used.

        """

        # Default fraction of max memory to be used
        fraction = params.get('memory', 0.8)

        # Get available virtual memory
        vmem = psutil.virtual_memory()

        max_vmem = int(fraction * vmem.available)

        # use AS as alternative to VMEM if the attribute isn't defined.
        # http://stackoverflow.com/a/30269998/5731870
        if hasattr(resource,'RLIMIT_VMEM'):
            resource.setrlimit(resource.RLIMIT_VMEM,(max_vmem,max_vmem))
        elif hasattr(resource,'RLIMIT_AS'):
            resource.setrlimit(resource.RLIMIT_AS, (max_vmem,max_vmem))

        soft, hard = resource.getrlimit(resource.RLIMIT_AS)
        soft = round(soft/(1.0*1024*1024))
        hard = round(hard/(1.0*1024*1024))

        return "Memory: Soft ({}MB) Hard ({}MB)".format(soft, hard)

    def get_metadata(self, filename=None,file_type=None):
        """
        Get reusable metadata dict with some standard fields.

        Parameters
        ---------
        filename: Name of the file covered by this metadata

        Returns
        -------
        metadata: Dictionary with a number of fields
        """
        metadata = OrderedDict()

        metadata["schema"] = "standalone:notebook"
        metadata["version"] = "1.0"
        metadata["timestamp"] = datetime.now().replace(microsecond=0).isoformat()
        metadata["platform"] = {
                'node': platform.node(),
                'os': platform.system(),
                'release': platform.release(),
                'processor': platform.processor(),
                'python': platform.python_version(),
                'distribution': distro.linux_distribution()
                }
        metadata["filename"] = filename
        metadata["filetype"] = mimetypes.guess_type(filename)[0]
        metadata["filesize"] = os.stat(filename).st_size
        if file_type in ["csv","tsv"]:
            if file_type == "csv":
                df = pd.read_csv(filename)
            elif file_type == "tsv":
                df = pd.read_csv(filename,sep="\t")
            metadata["rows"] = df.shape[0]
            metadata["columns"] = df.shape[1]
            metadata["headers"] = list(df.columns)

        return metadata


    def read_notebook(self, notebook_path,
                      as_version=nbformat.NO_CONVERT):
        """
        Parameters
        ----------
        notebook_path: Absolute path of ipynb file
        as_version: Version of Notebook. Default = nbformat.NO_CONVERT

        Returns
        -------
        NotebookNode: Dictionary representation of file
        """
        try:
            notebook = nbformat.read(notebook_path,
                                     as_version=as_version)
            return notebook
        except Exception as e:
            print("Exception reading Notebook - {}".format(e))

    def search_within_notebook(self,
                               notebook_path,
                               keyword):
        """
        Parameters
        ----------
        notebook_path: Absolute path of ipynb file
        keyword: String to search for within notebook

        Returns
        -------
        Boolean, true if keyword is found in ipynb file; False if not.
        """
        try:

            notebook = self.read_notebook(notebook_path)
            if keyword in notebook:
                return True
            return False
        except Exception as e:
            print("Exception searching in notebook - {}".format(e))

    def search_notebooks(self, user_notebook_dir, keyword):
        """
        Parameters
        ----------
        notebook_path: Absolute path of ipynb file
        keyword: String to search for within notebook

        Returns
        -------
        List of path strings to Notebooks that satisfy the search
        """
        try:
            notebooks = []
            ipynb_files = glob.glob(os.path.join(notebook_dir,
                                                 "/**/*.ipynb"),
                                    recursive=True)
            for ipynb in ipynb_files:
                present = self.search_within_notebook(ipynb, keyword)
                if present:
                    notebooks.append(ipynb)
            return notebooks

        except Exception as e:
            print("Exception searching in notebook directory - {}".format(e))

    def get_file(self, path, create_dir=False):
        """
        Parameters
        ----------
        path: Path specification
        create_dir: Should the parent director for the file be created

        Example
        ----------
        %(data_root)/acme/Projects/commands
        This is resolved into `/home/ubuntu/enrich/data/notebooks/acme/Projects/commands`

        Returns
        -------
        Full path from abstract specification
        """
        try:
            path = path.replace("%(",'%%(') # This will make the strftime handling safe
            path = datetime.now().strftime(path)
            path = path % {
                'data_root': self.data_root,
                'dt': datetime.now().date().isoformat(),
                'datetime': datetime.now().replace(microsecond=0).isoformat()
            }

            if create_dir:
                try:
                    os.makedirs(os.path.dirname(path))
                except:
                    pass

            return path
        except Exception as e:
            print("Exception in get_file() - {}".format(e))


    def save(self, file_name, metadata_path = None):
        """
        Parameters
        ----------
        file_name: Relative path of file including filename with extension

        Returns
        -------
        None. Saves the data file specified and metadata about the file into Enrich's data dir
        """
        try:
            metadata_file = None
            abs_file_name = self.get_file(file_name)
            file_type = abs_file_name.split(".")[-1]

            #Validate that file is "csv" or "tsv". More file types can be added to list
            if os.path.exists(abs_file_name) and file_type in accepted_types():
                metadata_info = self.get_metadata(abs_file_name,file_type)
                if metadata_path is None:
                    metadata_file = os.path.join(self.output_dir,'metadata.json')
                elif isinstance(metadata_path,str):
                    resolved_metadata_path = self.get_file(metadata_path)
                    create_dir('/'.join(resolved_metadata_path.split("/")[:-1]))
                    metadata_file = resolved_metadata_path
                elif callable(metadata_path):
                    data_dump_dir = metadata_path()

                with open(metadata_file, 'w') as f:
                    json.dump(metadata_info, f, indent=4)
        except Exception as e:
            print("Exception in save() - {}".format(e))



import os
import sys
import copy
import json
import re
from enrichsdk import Sink
import logging

logger = logging.getLogger("app")

class TableSink(Sink):
    """
    Transform to dump dataframes in state into files.

    Parameters specific to this module include:

        * sink: A dictionary of dataframe names and how to output them. It has a number of attributes:

            * type: Output type. Only 'table' value is supported for this
              option right now.
            * filename: Output filename. You can use default parameters such
              runid

          The name of the dataframe can be a regular expression allowing you
          specify a simple rule for arbitrary number of frames.

    Example::

        ....
        "transforms": {
            "enabled": [
                ...
                {
                    "transform": "TableSink",
                    "args": {
                        "sink": {
                            "article": {
                                "frametype": "pandas",
                                "filename": "%(output)s/%(runid)s/article.csv",
                                "params": {
                                    "sep": "|"
                                }
                            },
                            ...
                        }
                    }
                    ...
                }
            ]
         }
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "TableSink"

        self.testdata = {
            'conf': {
                'args': {
                    "sink": {
                        "cars": {
                            "frametype": "pandas",
                            "filename": "%(output)s/cars_revised.csv",
                            "params": {
                                "sep": ","
                            }
                        }
                    }
                }
	        },
            'data': {
                "carmodel": {
                    "transform": "CarModel",
                    "filename": "cars.csv",
                    "params": {
                        "sep": ","
                    },
                    "state": {
                        "params": [
                            {
                                'type': 'args',
                                'transform': 'TableSink',
                                'args': {
                                    'save': False,
                                    'rows': 124
                                }
                            }
                        ]
                    }
                }
            }
        }
    def preload_clean_args(self, args):
        """
        Check to make sure that the arguments
        is consistent with the specification mentioned
        above
        """

        # Sanity check...
        assert isinstance(args, dict)
        assert len(args) > 0

        # Backward compatability
        if 'sink' not in args:
            args = {
                'sink': args
            }

        args = super().preload_clean_args(args)

        assert 'sink' in args
        sink = args['sink']
        assert isinstance(sink, dict)
        assert len(sink) > 0

        for pattern, detail  in sink.items():

            if (("type" in detail) and
                ("frametype" not in detail) and
                (detail['type'] == 'table')):
                detail['frametype'] = 'pandas'

            if (("frametype" not in detail) or
                (detail['frametype'] != 'pandas')):
                logger.error("Invalid configuration. Only pandas table source supported by this sink transform",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            if (('filename' not in detail) or
                (not isinstance(detail['filename'], str)) or
                ('params' not in detail) or
                (not isinstance(detail['params'], dict))):
                logger.error("Invalid args. Filename (string) and params (dict) are required",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            detail['root'] = self.config.enrich_data_dir

            tags = detail.get('tags', [])
            if isinstance(tags, str):
                tags = [tags]
            detail['tags'] = tags

            sortcols = detail.get('sort', [])
            if isinstance(sortcols, str):
                sortcols = [sortcols]
            detail['sort'] = sortcols

        return args

    def validate_args(self, what, state):
        """
        Extra validation of the arguments
        """
        sink = self.args['sink']

        assert isinstance(sink, dict)
        for pattern, detail  in sink.items():
            assert (('frametype' in detail) and (detail['frametype'] == 'pandas'))
            assert 'filename' in detail
            assert 'params' in detail


    def process(self, state):
        """
        Execute the tablesink specification
        """

        # => What exists in the state?
        available_frames = state.get_frame_list()

        # => First construct input for the pandasframe
        extra = {}
        write_input = {}
        args_input = {}
        framecls = self.config.get_dataframe('pandas')
        skipped = []

        sink = self.args['sink']
        for pattern in sink:
            # The pattern could be precise dataframe name or could be
            # regular expression.
            regex = re.compile('^{}$'.format(pattern))
            frames = [m.group(0) for f in available_frames for m in [regex.search(f)] if m]

            for f in frames:
                # For each dataframe that is in the system

                detail = state.get_frame(f)

                # => Are there any extra instructions?
                overrides = self.frame_get_overrides(detail)

                #=> Materialize the path...
                filename = sink[pattern]['filename']
                filename = self.config.get_file(filename,
                                                create_dir=True,
                                                extra={
                                                    'frame_name': f
                                                })
                # Collect all column information
                extra[f] = {
                    'columns': self.collapse_columns(detail),
                    'notes': self.collapse_notes(detail),
                    'descriptions': self.collapse_descriptions(detail),
                    'overrides': overrides
                }

                # Which dataframe
                df = detail['df']

                # Get the
                frametype = detail['frametype']

                # Order the dataframe if it is needed
                sortcols = sink[pattern]['sort']
                if len(sortcols) > 0:
                    df = framecls.sort_values(df,
                                           sortcols,
                                           ascending=False)
                params = sink[pattern].get('params',{})

                # Should I be writing this csv?
                save = params.get('save', True)
                save = overrides.get('save', save)

                if not save:
                    skipped.append(f)

                write_input[f] = {
                    'save': save,
                    'frametype': frametype,
                    'pattern': pattern,
                    'df': df,
                    'filename': filename,
                    'params': params,
                }

                args_input[f] = copy.copy(sink[pattern])
                args_input[f]['filename'] = filename

        if len(skipped) > 0:
            logger.warning("Not saving {} tables".format(len(skipped)),
                           extra={
                               'transform': self.name,
                               'data': skipped
                           })

        # => Write output details
        framecls.write(args_input, write_input)

        for name in write_input:

            detail = write_input[name]

            # => Insert columns and tags
            pattern = detail['pattern']

            #
            detail['params']['tags'] = sink[pattern]['tags']

            # Incorporate columns, notes and description
            additional_params = extra[name]
            overrides = additional_params.pop('overrides', {})

            detail['params'].update(additional_params)

            # Insert any overrides provided in the state
            if 'rows' in overrides:
                detail['params']['components'][0]['rows'] = overrides['rows']

            detail['params'] = [
                detail['params'],
                {
                    "type": "lineage",
                    "transform": self.name,
                    "dependencies": [
                        {
                            "type": "dataframe",
                            "nature": "input",
                            "objects": [name]
                        },
                        {
                            "type": "file",
                            "nature": "output",
                            "objects": [detail['filename']]
                        }
                    ]
                }
            ]

            # Insert additional detail
            detail['transform'] = self.name
            detail['history'] = [
                {
                    'transform': self.name,
                    'log': "Wrote output"
                }
            ]

            state.update_frame(name, detail)

        logger.debug("Finished writing data",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

    def validate_results(self, what, state):
        pass


provider = TableSink



import os
import logging 
import click 
from logging.config import dictConfig 


def logging_start(filename=None): 

    default_filename = "ext://sys.stderr"
    if filename is None: 
        filename = default_filename 
    elif isinstance(filename, str): 
        try: 
            os.makedirs(os.path.dirname(filename)) 
        except: 

            pass 
        filename=open(filename, 'a')

    # logging.basicConfig(filename='example.log',level=logging.DEBUG)
    dictConfig({
        'version': 1, 
        "disable_existing_loggers": True,
        'formatters': {
            "basic": {
                "class": "logging.Formatter",
                "datefmt": "%I:%M:%S",
                "format": "%(asctime)s %(levelname)s %(name)s %(message)s"
            },
            'json': {
                '()': 'pythonjsonlogger.jsonlogger.JsonFormatter',
                'format': "%(asctime)s %(name)s %(levelname)s %(message)s",
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'basic',
                'level': 'DEBUG',        
                "stream": default_filename, 
            },
            'app': {
                'class': 'logging.StreamHandler',
                'formatter': 'json',
                'level': 'DEBUG',        
                "stream": filename, 
            }
        },
        'loggers': {
            '': {
                'level': 'DEBUG',
                'handlers': ['console'], 
                'propagate': False
            },
            'default': {
                'level': 'DEBUG',
                'handlers': ['console'],
                'propagate': False
            },
            'app': {
                'level': 'DEBUG',
                'handlers': ['app'],
                'propagate': False
            },
        },
        'root': {
            'level': 'DEBUG',
            #'handlers': ['app', 'console'],
            'handlers': ['app'],
        },
    })

def logging_end():
    logging.shutdown() 

def logging_clear(): 
    for h in logging._handlers.copy():
        logging.removeHandler(h)
        h.flush()
        h.close()



"""
spark.py
~~~~~~~~

Module containing helper function for use with Apache Spark
"""

import __main__

from os import environ, listdir, path
from json import loads

from pyspark import SparkFiles
from pyspark.sql import SparkSession

from dependencies import logging


def start_spark(app_name='my_spark_app', master='local[*]', jar_packages=[],
                files=[], spark_config={}):
    """Start Spark session, get Spark logger and load config files.

    Start a Spark session on the worker node and register the Spark
    application with the cluster. Note, that only the app_name argument
    will apply when this is called from a script sent to spark-submit.
    All other arguments exist solely for testing the script from within
    an interactive Python console.

    This function also looks for a file ending in 'config.json' that
    can be sent with the Spark job. If it is found, it is opened,
    the contents parsed (assuming it contains valid JSON for the ETL job
    configuration), into a dict of ETL job configuration parameters,
    which are returned as the last element in the tuple returned by
    this function. If the file cannot be found then the return tuple
    only contains the Spark session and Spark logger objects and None
    for config.

    The function checks the enclosing environment to see if it is being
    run from inside an interactive console session or from an
    environment which has a `DEBUG` environment varibale set (e.g.
    setting `DEBUG=1` as an environment variable as part of a debug
    configuration within an IDE such as Visual Studio Code or PyCharm in
    In this scenario, the function uses all available function arguments
    to start a PySpark driver from the local PySpark package as opposed
    to using the spark-submit and Spark cluster defaults. This will also
    use local module imports, as opposed to those in the zip archive
    sent to spark via the --py-files flag in spark-submit.

    :param app_name: Name of Spark app.
    :param master: Cluster connection details (defaults to local[*].
    :param jar_packages: List of Spark JAR package names.
    :param files: List of files to send to Spark cluster (master and
        workers).
    :param spark_config: Dictionary of config key-value pairs.
    :return: A tuple of references to the Spark session, logger and
        config dict (only if available).
    """

    # detect execution environment
    flag_repl = False if hasattr(__main__, '__file__') else True
    flag_debug = True if 'DEBUG' in environ.keys() else False

    print("MASTER", master)

    masterset = master != 'local[*]'
    if not (flag_repl or flag_debug or masterset):
        # get Spark session factory
        spark_builder = (
            SparkSession
            .builder
            .master(master)            
            .appName(app_name))
    else:
        # get Spark session factory
        spark_builder = (
            SparkSession
            .builder
            .master(master)
            .appName(app_name))

        # create Spark JAR packages string
    if len(jar_packages) > 0: 
        spark_jars_packages = ','.join(list(jar_packages))
        spark_builder.config('spark.jars.packages', spark_jars_packages)

    if len(files) > 0: 
        spark_files = ','.join(list(files))
        spark_builder.config('spark.files', spark_files)

    # add other config params
    for key, val in spark_config.items():
        spark_builder.config(key, val)

    # create session and retrieve Spark logger object
    spark_sess = spark_builder.getOrCreate()
    spark_logger = logging.Log4j(spark_sess)

    spark_logger.info("Master: {}".format(master))
    
    # get config file if sent to cluster with --files
    spark_files_dir = SparkFiles.getRootDirectory()
    config_files = [filename
                    for filename in listdir(spark_files_dir)
                    if filename.endswith('config.json')]

    if len(config_files) != 0:
        path_to_config_file = path.join(spark_files_dir, config_files[0])
        with open(path_to_config_file, 'r') as config_file:
            config_json = config_file.read().replace('\n', '')
        config_dict = loads(config_json)
        spark_logger.warn('loaded config from ' + config_files[0])
    else:
        spark_logger.warn('no config file found')
        config_dict = None

    return spark_sess, spark_logger, config_dict



import os
import json
import logging
from collections import defaultdict
from datetime import datetime, date, timedelta
from dateutil import parser as dateparser
import pandas as pd

from enrichsdk import Compute, S3Mixin, CheckpointMixin
from enrichsdk.feature_compute import *

logger = logging.getLogger("app")

__all__ = [
    'FeatureComputeBase',
]


class FeatureComputeBase(Compute):
    """
    A built-in transform baseclass to handle standard feature
    computation and reduce the duplication of code.

    This should be used in conjunction with an FeaturesetExtractor & FeatureExtractor
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "FeatureComputeBase"
        self._environ = os.environ.copy()

    @classmethod
    def instantiable(cls):
        """
        Return true if class can be instantiated. Override in subclass
        """
        return False

    def get_featureset_extractors(self):
        """
        Get all the *featureset* extractors (not feature extractors)

        Returns:
             list: A list of extractors as a name, extractor combination

        For example::

            return [{
                 "name": "patient",
                 "extractor": <featureset extractor instance>
            }]

        """
        raise Exception("Implement in subclass")

    def store(self, data):
        """
        Store the final result

        Args:
            data (dict): name of featureset -> data associated with it

        """
        raise Exception("Implement in subclass")

    def get_objects(self):
        """
        Get a list of objects (typically names)  to process. Could be dictionaries,
        lists etc. The list is not interpreted by the base class. Could be a list of
        identifier.

        Returns:
           list: A list of objects (could be ids/paths/dicts etc.)

        """
        if 'root' not in args:
            raise Exception("Base class implementation required 'root'")

        root = self.args['root']
        files = os.listdir(root)
        return files

    def read_object(self, obj):
        """
        Read one object returned by get_objects

        Args:
            obj (object): One item in the list of objects

        Returns:
             object: An object like dict or list of dicts

        """

        if 'root' not in args:
            raise Exception("Base class implementation required 'root'")

        root = self.args['root']
        filename = os.path.join(root, obj)
        data = json.load(open(filename))
        return data

    def process(self, state):
        """
        Core loop

        Rough logic::

            get featureset extractors
            get objects
            for each object:
                 for each featureset extractor X
                     process one object with X
                     collect one featureset 'row' for X

            for each featureset extractor X

        """

        logger.debug("Start execution",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

        self.state = state

        # What extractors to run on the data..
        featureset_extractors = self.get_featureset_extractors()
        featureset_extractors = [f for f in featureset_extractors if f.get('enable', True)]

        # Go through all the available objects
        objects = self.get_objects()
        logger.debug(f"Received {len(objects)} objects",
                     extra={
                         'transform': self.name
                     })

        # Compute the features..
        final = compute_features(objects,
                                 featureset_extractors,
                                 self.read_object)

        # Update the frame
        for name, df in final.items():
             if isinstance(df, pd.DataFrame):
                self.update_frame(name + "_features",
                                  "Features computed over the available data",
                                  df, objects[0])

        # Store the result...
        self.store(final)

        logger.debug("Complete execution",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass



import psutil
import resource
import logging

logger = logging.getLogger('app')

def set_limits_memory(params={}):

    # Default fraction of max memory to be used 
    fraction = params.get('memory', 0.8)
    
    # Get available virtual memory 
    vmem = psutil.virtual_memory()

    max_vmem = int(fraction * vmem.available)
    
    # use AS as alternative to VMEM if the attribute isn't defined.
    # http://stackoverflow.com/a/30269998/5731870
    if hasattr(resource,'RLIMIT_VMEM'):
        resource.setrlimit(resource.RLIMIT_VMEM,(max_vmem,max_vmem))
    elif hasattr(resource,'RLIMIT_AS'):
        resource.setrlimit(resource.RLIMIT_AS, (max_vmem,max_vmem))

    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    soft = round(soft/(1.0*1024*1024))
    hard = round(hard/(1.0*1024*1024))
    
    return "Memory: Soft ({}MB) Hard ({}MB)".format(soft, hard)

def set_limits(params={}):
    msg = set_limits_memory(params)

    logger.debug("Resource Limits set",
                 extra={
                     'transform': 'Enricher',
                     'data': msg
                 })
